'''
Module docstring
'''

import sys
import os
from importlib import reload
import tempfile
import string
import pyutilib
import contextlib
from collections import namedtuple
import numbers

import numpy as np
import pandas as pd

import pyomo.environ as po
from pyomo.core.base.objective import SimpleObjective
from pyomo.opt import SolverFactory

import grimsel.auxiliary.maps as maps
import grimsel.auxiliary.timemap as timemap

import grimsel.core.constraints as constraints
import grimsel.core.variables as variables
import grimsel.core.parameters as parameters
import grimsel.core.sets as sets
import grimsel.core.io as io # for class methods
from grimsel import _get_logger

logger = _get_logger(__name__)


def get_random_suffix():
    return ''.join(np.random.choice(list(string.ascii_lowercase), 4))
#
#TEMP_DIR = 'grimsel_temp_' + get_random_suffix()

#def create_tempfile(self, suffix=None, prefix=None, text=False, dirc=None):
#    """
#    Return the absolute path of a temporary filename that is_init_pf_dicts
#    guaranteed to be unique.  This function generates the file and returns
#    the filename.
#    """
#
#    logger.warn('!!!!!!!!tempfiles.TempfileManagerPlugin.create_tempfile '
#                'is monkey patched!!!!!!!!')
#
#    if suffix is None:
#        suffix = ''
#    if prefix is None:
#        prefix = 'tmp'
#
#    ans = tempfile.mkstemp(suffix=suffix, prefix=prefix, text=text, dir=dirc)
#    ans = list(ans)
#    if not os.path.isabs(ans[1]):  #pragma:nocover
#        fname = os.path.join(dirc, ans[1])
#    else:
#        fname = ans[1]
#    os.close(ans[0])
#
#    dirc = TEMP_DIR
#
#    if not os.path.isdir(dirc):
#        os.mkdir(dirc)
#
#    new_fname = os.path.join(dirc, 'grimsel_temp_' + get_random_suffix() + suffix)
#    # Delete any file having the sequential name and then
#    # rename
#    if os.path.exists(new_fname):
#        os.remove(new_fname)
#    fname = new_fname
#
#    self._tempfiles[-1].append(fname)
#    return fname
#
#import pyutilib.component.config.tempfiles as tempfiles
#tempfiles.TempfileManagerPlugin.create_tempfile = create_tempfile

reload(constraints)
reload(variables)
reload(parameters)
reload(sets)

class ModelBase(po.ConcreteModel, constraints.Constraints,
                parameters.Parameters, variables.Variables, sets.Sets):

    # class attributes as defaults for presolve_fixed_capacities
    list_vars = [('var_yr_cap_pwr_rem', 'cap_pwr_rem'),
                 ('var_yr_cap_pwr_new', 'cap_pwr_new')]
    list_constr_deact = ['set_win_sol']

#    db = get_config('sql_connect')['db']


    def __init__(self, **kwargs):
        '''
        Keyword arguments:
        nhours -- time resolution of the model, used for profile scaling
        sc_warmstart -- input database schema for presolving
        slct_node -- limit node selection
        slct_encar -- limit to energy carrier selection
        skip_runs -- boolean; if True, solver calls are skipped, also
                     stops the IO instance from trying to write the model
                     variables.
        '''

        super(ModelBase, self).__init__() # init of po.ConcreteModel

        defaults = {'slct_node': [],
                    'slct_pp_type': [],
                    'slct_node_connect': [],
                    'slct_encar': [],
                    'nhours': 1,
                    'unq_code': '',
                    'mps': None,
                    'tm_filt': False,
                    'verbose_solver': True,
                    'constraint_groups': None,
                    'symbolic_solver_labels': False,
                    'skip_runs': False,
                    'nthreads': False,
                    'keepfiles': True,
                    'tempdir': None}
        for key, val in defaults.items():
            setattr(self, key, val)
        self.__dict__.update(kwargs)

        self._check_contraint_groups()

        logger.info('self.slct_encar=' + str(self.slct_encar))
        logger.info('self.slct_pp_type=' + str(self.slct_pp_type))
        logger.info('self.slct_node=' + str(self.slct_node))
        logger.info('self.slct_node_connect=' + str(self.slct_node_connect))
        logger.info('self.nhours=' + str(self.nhours))
        logger.info('self.constraint_groups=' + str(self.constraint_groups))

        self.warmstartfile = self.solutionfile = None

        # attributes for presolve_fixed_capacities
        self.list_vars = ModelBase.list_vars
        self.list_constr_deact = ModelBase.list_constr_deact


#    def _update_slct_lists(self):
#        ''''''
#        for attr, series in ((self.slct_encar, self.df_def_encar.ca),
#                             (self.slct_encar_id, self.df_def_encar.ca_id),
#                             (self.slct_pp_type_id, self.df_def_pp_type.pt),
#                             (self.slct_pp_type_id, self.df_def_pp_type.pt),
#                             (self.slct_node, self.df_def_node.nd)):
#            if not attr:
#                attr = series.tolist()


    def build_model(self):
        '''
        Call the relevant model methods to get everything set up.

        This consists in:
        1. call self.get_setlst (in Sets mixin class) to initialize
            the self.setlst dictionary
        2. call self.define_sets (in Sets mixin class) to initialize
            Pyomo set objects
        3. call self.define_parameters (in Parameters mixin class)
        4. call self.define_variables (in Variables mixin class)
        5. call self.add_all_constraints
        6. call self.init_solver

        .. note::
           io needs to have loaded all data, i.e. set the ModelBase
           DataFrames.
        '''

        self.get_setlst()
        self.define_sets()

        self.define_parameters()

        if not self.skip_runs:

            self.define_variables()
            self.add_all_constraints()
            self.init_solver()



    @classmethod
    def get_constraint_groups(cls, excl=None):
        '''
        Returns list names of methods defining constraint groups.

        This classmethod can also be used to define the constraint_groups
        parameter to initialize the ModelBase object by selecting certain
        groups to be excluded.

        Parameters
        ----------
        excl : list
            exclude certain group names from the returned list

        Returns
        -------
        list
            Names of constraint groups. These correspond to the
            methods in the :class:`Constraints` class without the prefix
            `add_` and the suffix `_rules`

        '''

        cg_lst = [mth.replace('add_', '').replace('_rules', '')
                  for mth in dir(cls)
                  if mth.startswith('add_') and 'rule' in mth]

        if excl:
            cg_lst = [cg for cg in cg_lst if not cg in excl]

        return cg_lst


    def _check_contraint_groups(self):
        '''
        Verification and completion of the constraint group selection.

        Verifies constraint groups if the ``constraint_groups`` argument
        is not None. Otherwise it gathers all accordingly named
        methods from the class attributes and populates the list thusly.

        Raises
        ------
        ValueError
            If the :class:`ModelBase` instance attribute ``constraint_groups``
            contains invalid entries.

        '''

        cg_options = self.get_constraint_groups()

        if self.constraint_groups is None:
            self.constraint_groups = cg_options
        else:
            # get invalid constraint groups in input
            nv = [cg for cg in self.constraint_groups
                  if not cg in cg_options]

            if nv:
                estr = ('Invalid constraint group(s): {nv}.'
                        + '\nPossible choices are:\n{cg}'
                        ).format(nv=', '.join(nv), cg=',\n'.join(cg_options))
                raise ValueError(estr)

    def add_all_constraints(self):
        '''
        Call all selected methods from the constraint mixin class.

        Loops through the `constraint_groups` list and calls the corresponding
        methods in the :class:`.Constraints` mixing class.
        '''

        for cg in set(self.constraint_groups):
            logger.info('##### Calling constraint group {}'.format(cg.upper()))
            getattr(self, 'add_%s_rules'%cg)()


    def _limit_prof_to_cap(self):

        if len(self.chp) > 0:
            self.limit_prof_to_cap()


    def limit_prof_to_cap(self, param_mod='cap_pwr_leg'):
        '''
        Make sure CHP profiles don't ask for more power than feasible.

        This operates on the parameters and is called before each model run.
        '''

        logger.info('Limiting chp profiles to cap_pwr_leg')

        # get list of plants relevant for chp from corresponding set
        pp_chp = self.setlst['chp']

        df_chpprof = io.IO.param_to_df(self.chpprof, ('sy', 'nd_id', 'ca_id'))
        df_erg_chp = io.IO.param_to_df(self.erg_chp, ('pp_id', 'ca_id'))
        df_erg_chp = df_erg_chp.loc[df_erg_chp.pp_id.isin(pp_chp)]
        df_erg_chp['nd_id'] = df_erg_chp.pp_id.replace(self.mps.dict_plant_2_node_id)

        # outer join profiles and energy to get a profile for each fuel
        df_chpprof_tot = pd.merge(df_erg_chp.rename(columns={'value': 'erg'}),
                                  df_chpprof.rename(columns={'value': 'prof'}),
                                  on=['nd_id', 'ca_id'])
        # scale profiles
        df_chpprof_tot['prof_sc'] = df_chpprof_tot['erg'] * df_chpprof_tot['prof']

        # get capacities from parameter
        df_cap_pwr_leg = io.IO.param_to_df(self.cap_pwr_leg, ('pp_id', 'ca_id'))
        # keep only chp-related fuels
        df_cap_pwr_leg = df_cap_pwr_leg.loc[df_cap_pwr_leg.pp_id.isin(self.chp)]
        # pivot_by fl_id
        df_cappv = df_cap_pwr_leg.pivot_table(values='value',
                                              index=['ca_id', 'pp_id'],
                                              aggfunc=np.sum)['value']
        # rename
        df_cappv = df_cappv.rename('cap').reset_index()

        # add capacity to profiles
        df_chpprof_tot = pd.merge(df_cappv, df_chpprof_tot, on=['ca_id', 'pp_id'])


        # find occurrences of capacity zero and chp erg non-zero
        df_slct = df_chpprof_tot[['pp_id', 'ca_id', 'cap', 'erg']].drop_duplicates().copy()
        df_slct = df_slct.loc[df_slct.cap.isin([0])
                            & -df_slct.erg.isin([0])]
        str_erg_cap = ''
        if len(df_slct > 0):
            for nrow, row in df_slct.iterrows():
                str_erg_cap += 'pp_id=%d, ca_id=%d: cap_pwr_leg=%f, erg_chp=%f\n'%tuple(row.values)
            raise ValueError ('limit_prof_to_cap: one or more cap_pwr_leg are zero '
                              'while erg_chp is greater 0: \n' + str_erg_cap)

        # find occurrences of capacity violations
        mask_viol = df_chpprof_tot.prof_sc > df_chpprof_tot.cap


        if mask_viol.sum() == 0:
            logger.info('ok, nothing changed.')
        else:
            # REPORTING
            df_profviol = df_chpprof_tot.loc[mask_viol]
            dict_viol = df_profviol.pivot_table(index=['pp_id', 'ca_id'],
                                                values='sy', aggfunc=len)['sy'].to_dict()

            for kk, vv in dict_viol.items():
                logger.warning('limit_prof_to_cap: \n(pp, ca)='
                               '{}: {} violations'.format(kk, vv))

            logger.warning('limit_prof_to_cap: Modifing model '
                           'parameter ' + param_mod)

            if param_mod == 'chpprof':

                df_profviol['prof'] *=  0.999 * df_chpprof_tot.cap / df_chpprof_tot.prof_sc
                dict_chpprof = (df_profviol.pivot_table(index=['sy', 'nd_id', 'ca_id'],
                                                        values='prof', aggfunc=min)['prof']
                                           .to_dict())

                for kk, vv in dict_chpprof.items():
                    self.chpprof[kk] = vv

            elif param_mod == 'cap_pwr_leg':

                # calculate capacity scaling factor
                df_capsc = df_profviol.pivot_table(index=['pp_id', 'ca_id'],
                                                   values=['cap', 'prof_sc'], aggfunc=np.max)
                df_capsc['cap_sc'] = df_capsc.prof_sc / df_capsc.cap

                # merge scaling factor with capacity table
                df_cap_pwr_leg = df_cap_pwr_leg.join(df_capsc,
                                                     on=df_capsc.index.names)
                df_cap_pwr_leg = df_cap_pwr_leg.loc[-df_cap_pwr_leg.cap_sc.isnull()]

                # apply scaling factor to all capacity with the relevant fuel
                df_cap_pwr_leg['cap'] *= df_cap_pwr_leg.cap_sc * 1.0001

                # dictionary
                dict_cap_pwr_leg = df_cap_pwr_leg.set_index(['pp_id', 'ca_id'])['cap']
                dict_cap_pwr_leg = dict_cap_pwr_leg.to_dict()

                for kk, vv in dict_cap_pwr_leg.items():
                    self.cap_pwr_leg[kk] = vv


    def _init_pf_dicts(self):
        '''
        Initializes dicts mapping the profile ids to other model ids.

        This results in dictionaries which are assigned as :class:`ModelBase`
        instance attributes:

        * ``dict_pricesll_pf``: (fl_id, nd_id, ca_id) |rarr| (pricesll_pf_id)
        * ``dict_pricebuy_pf``: (fl_id, nd_id, ca_id) |rarr| (pricebuy_pf_id)
        * ``dict_dmnd_pf``: (nd_id, ca_id) |rarr| (dmnd_pf_id)
        * ``dict_supply_pf``: (pp_id, ca_id) |rarr| (supply_pf_id)

        Purpose
        ---------

        The resulting dictionaries are used for filtering the profile tables
        in the :module:`io` module and to access the profile parameters
        in the model :class:`Constraints`.

        '''

        list_pf = [(self.df_fuel_node_encar,
                    ['fl_id', 'nd_id', 'ca_id'], 'pricebuy'),
                   (self.df_fuel_node_encar,
                    ['fl_id', 'nd_id', 'ca_id'], 'pricesll'),
                   (self.df_node_encar,
                    ['nd_id', 'ca_id'], 'dmnd'),
                   (self.df_plant_encar,
                    ['pp_id', 'ca_id'], 'supply')]

        df, ind, name = list_pf[-1]
        for df, ind, name in list_pf:

            col = '%s_pf_id'%name
            if df is not None and col in df.columns:
                ind_df = df.loc[~df[col].isna()].set_index(ind)[col]
                dct = ind_df.to_dict()
            else:
                dct = {}

            setattr(self, 'dict_%s_pf'%name, dct)

    def translate_pf_id(self, df):
        '''
        Adds model id columns for the profile ids in the input DataFrame.

        Searches vars(self) for the pf_dict corresponding to the pf_ids
        in the input DataFrame. Then uses this dictionary to add additional
        columns to the output table.

        Parameters
        ----------
        df (DataFrame): DataFrame with pf_id column.

        Returns
        -------
        :obj:`pandas.DataFrame`
            Input DataFrame with added model ids corresponding to the pf_id.

        Raises
        ------
        IndexError: If multiple pf dictionaries correspond to the pf_id
                    values in the input DataFrame.
        IndexError: If no pf dictionary can be found for the pf_id values.

        '''

        # identify corresponding pf dict
        list_pf_id = set(df.pf_id.unique().tolist())

        pf_arrs = {name_dict:
                    list_pf_id
                        .issubset(set(getattr(self, name_dict).values()))
                    for name_dict in vars(self)
                    if name_dict.startswith('dict_')
                    and name_dict.endswith('_pf')}

        if sum(pf_arrs.values()) > 1:
            raise ValueError('Ambiguous pf array in translate_pf_id '
                             'or df empty.')
        elif sum(pf_arrs.values()) == 0:
            raise ValueError('No pf array found for table with columns '
                             '%s'%df.columns.tolist() + '. Maybe you are '
                             'trying to translate a table with pf_ids which '
                             'are not included in the original model.')
        else:
            pf_dict = {val: key for key, val in pf_arrs.items()}[True]

            new_cols = {'dict_pricesll_pf': ['fl_id', 'nd_id', 'ca_id'],
                        'dict_pricebuy_pf': ['fl_id', 'nd_id', 'ca_id'],
                        'dict_price_pf': ['fl_id', 'nd_id', 'ca_id'],
                        'dict_dmnd_pf': ['nd_id', 'ca_id'],
                        'dict_supply_pf': ['pp_id', 'ca_id']}[pf_dict]

            pf_dict = getattr(self, pf_dict)


            df_new = pd.Series(pf_dict).reset_index()
            df_new.columns = new_cols + ['pf_id']

            df_new = pd.merge(df_new, df, on='pf_id')

            return df_new


    def _get_nhours_nodes(self, nhours):
        '''
        Generates the nhours dictionary ``nhours``.

        Returns
        -------
        nhours_dict : dict
            ``{node_1: (original time res, target time res),
               node_2: (original time res, target time res),
               ...}``

        '''

        logger.debug('Generating nhours dictionary for '
                    'nhours={} with type {}'.format(nhours, type(nhours)))

        if isinstance(nhours, dict):

            nhours_dict = {}

            for nd_id in self.slct_node_id:

                nd = self.mps.dict_nd[nd_id]

                if nd in nhours:

                    if isinstance(nhours[nd], tuple):
                        # all there
                        nhours_dict[nd_id] = nhours[nd]

                    elif isinstance(nhours[nd], numbers.Number):
                        # assuming original time resolution 1 hour
                        nhours_dict[nd_id] = (1, nhours[nd])

                else:
                    # assuming default
                    nhours_dict[nd_id] = (1, 1)

        elif isinstance(nhours, numbers.Number):
            nhours_dict = {nd: (1, nhours) for nd in self.slct_node_id}
        elif isinstance(nhours, tuple):
            nhours_dict = {nd: nhours for nd in self.slct_node_id}
        else:
            raise ValueError(f'Unknown structure of `nhours`: {nhours}.\n'
                               f'Must be one of\n'
                               f'* number: constant target time resolution,'
                               f'  assuming 1 hour input data time resolution\n'
                               f'* tuple (input_t_res, target_t_res): applied'
                               f'  to all nodes\n'
                               '* dict {nd: (input, target)}: explicit '
                                'specification')

        return nhours_dict


    def init_maps(self):
        '''
        Uses the input DataFrames to initialize a
        :class:`grimsel.auxiliary.maps.Maps` instance.

        '''

        dct = {var.replace('df_def_', ''): getattr(self, var)
               for var in vars(self) if 'df_def_' in var}

        self.mps = maps.Maps.from_dicts(dct)

    def _init_time_map_connect(self):

        df_ndcnn = self.df_node_connect[['nd_id', 'nd_2_id', 'ca_id']].drop_duplicates()

        df_ndcnn['freq'] = df_ndcnn.nd_id.apply(lambda x: {key: frnh[0] for key, frnh in self._dict_nd_tm.items()}[x])
        df_ndcnn['nhours'] = df_ndcnn.nd_id.apply(lambda x: {key: frnh[1] for key, frnh in self._dict_nd_tm.items()}[x])
        df_ndcnn['freq_2'] = df_ndcnn.nd_2_id.apply(lambda x: {key: frnh[0] for key, frnh in self._dict_nd_tm.items()}[x])
        df_ndcnn['nhours_2'] = df_ndcnn.nd_2_id.apply(lambda x: {key: frnh[1] for key, frnh in self._dict_nd_tm.items()}[x])
        df_ndcnn['tm_id'] = df_ndcnn.nd_id.replace(self.dict_nd_tm_id)
        df_ndcnn['tm_2_id'] = df_ndcnn.nd_2_id.replace(self.dict_nd_tm_id)

        # make dict_sy_ndnd_min
        is_min_node = pd.concat([df_ndcnn,
                                 df_ndcnn.assign(nd_id = df_ndcnn.nd_2_id,
                                                 nd_2_id = df_ndcnn.nd_id,
                                                 nhours = df_ndcnn.nhours_2,
                                                 nhours_2 = df_ndcnn.nhours)])
        self.is_min_node = (
                is_min_node.assign(is_min=is_min_node.nhours
                                   <= is_min_node.nhours_2)
                                   .set_index(['nd_id', 'nd_2_id'])
                                   .is_min).to_dict()

        def get_map_sy(x):
            tm = timemap.TimeMap(tm_filt=self.tm_filt, minimum=True,
                                 freq=x[['freq', 'freq_2']].min(axis=1).values[0],
                                 nhours=x.nhours.iloc[0])
            tm_2 = timemap.TimeMap(tm_filt=self.tm_filt, minimum=True,
                                   freq=x[['freq', 'freq_2']].min(axis=1).values[0],
                                   nhours=x.nhours_2.iloc[0])
            return pd.merge(tm.df_hoy_soy,
                            tm_2.df_hoy_soy.rename(columns={'sy': 'sy2'}),
                            on='hy')[['sy', 'sy2']]

        self.dict_ndnd_tm_id =  df_ndcnn.set_index(['nd_id', 'nd_2_id']).copy()
        self.dict_ndnd_tm_id['tm_min_id'] = self.dict_ndnd_tm_id.apply(lambda x: x.tm_id if x.nhours <= x.nhours_2 else x.tm_2_id, axis=1)
        self.dict_ndnd_tm_id = self.dict_ndnd_tm_id.tm_min_id.to_dict()
        self.dict_ndnd_tm_id = {**self.dict_ndnd_tm_id,
                                **{(key[1], key[0]): val
                                   for key, val
                                   in self.dict_ndnd_tm_id.items()}}

        sysymap = df_ndcnn[[c for c in df_ndcnn.columns if not 'nd' in c]]
        sysymap = df_ndcnn.drop_duplicates()
        sysymap = df_ndcnn.groupby(['tm_id', 'tm_2_id']).apply(get_map_sy).reset_index()
        sysymap = sysymap.drop('level_2', axis=1)

        self.df_sysy_ndcnn = pd.merge(
                df_ndcnn[['nd_id', 'nd_2_id', 'ca_id', 'tm_id', 'tm_2_id']],
                sysymap, on=['tm_id', 'tm_2_id'], how='outer')

        self.dict_sysy = {**self.df_sysy_ndcnn.groupby(['nd_id', 'nd_2_id', 'sy']).sy2.apply(lambda x: set((*x,))).to_dict(),
                          **self.df_sysy_ndcnn.groupby(['nd_2_id', 'nd_id', 'sy2']).sy.apply(lambda x: set((*x,))).to_dict()}

        self.df_symin_ndcnn = self.df_sysy_ndcnn.join(df_ndcnn.set_index(['nd_id', 'nd_2_id'])[['nhours', 'nhours_2']], on=['nd_id', 'nd_2_id'])

        idx = ['nd_id', 'nd_2_id']
        cols = ['sy', 'sy2', 'tm_2_id', 'ca_id', 'tm_id', 'nhours', 'nhours_2']
        list_df = []
        for nd_id, nd_2_id in set(self.df_symin_ndcnn.set_index(idx).index.values):
            df = self.df_symin_ndcnn.set_index(idx).loc[[(nd_id, nd_2_id)], cols]

            nd_smaller = df.nhours.iloc[0] <= df.nhours_2.iloc[0]
            list_df.append(df.assign(symin = df.sy if nd_smaller else df.sy2,
                                     tm_min_id = df.tm_id if nd_smaller
                                                 else df.tm_2_id))

        cols = ['tm_min_id', 'symin', 'nd_id', 'nd_2_id', 'ca_id']
        self.df_symin_ndcnn = pd.concat(list_df).reset_index()[cols]

    def _init_time_map_input(self):
        '''
        If a *tm_soy* table is provided in the input data, time slots
        are assumed to be exogenously defined.

        '''

        # assert number of time slots in all profiles equals time slots in
        # tm_soy input table
        nsy = len(self.df_tm_soy.sy)
        assert(len(self.df_profdmnd.hy.unique()) == nsy), \
            'Inconsistent time slots tm_soy/profsupply'
        if getattr(self, 'profsupply', None) is not None:
            assert(len(self.df_profsupply.hy.unique()) == nsy), \
                'Inconsistent time slots tm_soy/profsupply'
        if getattr(self, 'profchp', None) is not None:
            assert(len(self.df_profchp.hy.unique()) == nsy), \
                'Inconsistent time slots tm_soy/profchp'
        if getattr(self, 'profchp', None) is not None:
            assert(len(self.df_profinflow.hy.unique()) == nsy), \
                'Inconsistent time slots tm_soy/profinflow'

        assert(len(self.df_tm_soy.weight.unique()) == 1), \
            'Multiple weights in input df_tm_soy. Can\'t infer time resolution.'

        # nhours follows from weight in df_tm_soy
        self.nhours = self.df_tm_soy.weight.iloc[0]
        self._dict_nd_tm = self._get_nhours_nodes(self.nhours)

        # generate unique time map ids
        dict_tm = {ntm: frnh for ntm, frnh
                   in enumerate(set(self._dict_nd_tm.values()))}

        self.dict_nd_tm_id = {nd:
                              {val: key for key, val in dict_tm.items()}[tm]
                              for nd, tm in self._dict_nd_tm.items()}

        self._tm_objs = {}

        self.df_def_node['tm_id'] = (self.df_def_node.reset_index().nd_id
                                         .replace(self.dict_nd_tm_id).values)

        unique_tm_id = self.df_def_node['tm_id'].iloc[0]

        self.df_hoy_soy = self.df_tm_soy[['sy']].assign(
                hy=lambda x: x.sy, tm_id=unique_tm_id)

        self.df_tm_soy = self.df_tm_soy.assign(tm_id=unique_tm_id,
                                               mt_id=None)

        self.df_tm_soy_full = None
        self.dict_soy_week = None
        self.dict_soy_month = None
        self.dict_week_soy = None
        self.dict_month_soy = None
        self.df_sy_min_all = None

        # dict pp_id -> tm
        self.dict_pp_tm_id = (
            self.df_def_plant.assign(tm_id=self.df_def_plant.nd_id
                                               .replace(self.dict_nd_tm_id))
                             .set_index('pp_id').tm_id.to_dict())

        # dict tm_id -> sy
        unq_list = lambda x: list(set(x))
        pv_kws = dict(index='tm_id', values='sy', aggfunc=unq_list)
        self.dict_tm_sy = self.df_hoy_soy.pivot_table(**pv_kws).sy.to_dict()





    def _init_time_map(self):
        '''
        Create a TimeMap instance and obtain derived attributes.

        Generated attributes:
            * ``tm`` (``TimeMap``): TimeMap object
            * ``df_tm_soy`` (``DataFrame``): timemap table
            * ``df_tm_soy_full`` (``DataFrame``): full timemap table

        '''

        self._dict_nd_tm = self._get_nhours_nodes(self.nhours)

        # generate unique time map ids
        dict_tm = {ntm: frnh for ntm, frnh
                   in enumerate(set(self._dict_nd_tm.values()))}

        self.dict_nd_tm_id = {nd:
                              {val: key for key, val in dict_tm.items()}[tm]
                              for nd, tm in self._dict_nd_tm.items()}

        self._tm_objs = {tm_id:
                   timemap.TimeMap(tm_filt=self.tm_filt,
                                   nhours=frnh[1], freq=frnh[0])
                   for tm_id, frnh in dict_tm.items()}

        self.df_def_node['tm_id'] = (self.df_def_node.reset_index().nd_id
                                         .replace(self.dict_nd_tm_id).values)

        cols_red = ['wk_id', 'mt_id', 'sy', 'weight', 'wk_weight']

        list_tm_soy = [tm.df_time_red[cols_red].assign(tm_id=tm_id) for
                       tm_id, tm in self._tm_objs.items()]
        self.df_tm_soy = pd.concat(list_tm_soy, axis=0)

        list_tm_soy_full = [tm.df_time_red.assign(tm_id=tm_id)
                            for tm_id, tm in self._tm_objs.items()]
        self.df_tm_soy_full = pd.concat(list_tm_soy_full, axis=0, sort=True)

        list_hoy_soy = [tm.df_hoy_soy.assign(tm_id=tm_id)
                        for tm_id, tm in self._tm_objs.items()]
        self.df_hoy_soy = pd.concat(list_hoy_soy, axis=0)

        df = self.df_tm_soy

        # get dictionaries month/week <-> time slots;
        # these are used in the constraint definitions
        cl, nm = ('wk_id', 'week')
        for cl, nm in [('wk_id', 'week'), ('mt_id', 'month')]:

            dct = df.pivot_table(index=['tm_id', cl],
                                 values='sy', aggfunc=list).sy.to_dict()
            setattr(self, 'dict_' + nm + '_soy', dct)

            dct = df.set_index(['tm_id', 'sy'])[cl].to_dict()
            setattr(self, 'dict_soy_' + nm, dct)


        # dict pp_id -> tm
        self.dict_pp_tm_id = (
            self.df_def_plant.assign(tm_id=self.df_def_plant.nd_id
                                               .replace(self.dict_nd_tm_id))
                             .set_index('pp_id').tm_id.to_dict())


        # dict tm_id -> sy
        unq_list = lambda x: list(set(x))
        pv_kws = dict(index='tm_id', values='sy', aggfunc=unq_list)
        self.dict_tm_sy = self.df_hoy_soy.pivot_table(**pv_kws).sy.to_dict()


        self._make_minimum_time_map()


    def _make_minimum_time_map(self):
        '''
        Generate table mapping all time maps to the minimum time map.

        This is mainly used in the post-analysis to map all data to uniform
        time slots.

        Example
        =======

        In the example below the minimum time map ``tm_id == 2`` indexes
        the table ``self.df_symin_all``

        ======= ======= =======
        sy_min  tm_id   sy
        ======= ======= =======
        0       0       0
        1       0       1
        2       0       2
        3       0       3
        4       0       4
        0       1       0
        1       1       0
        2       1       1
        3       1       1
        ...     ...     ...
        ======= ======= =======

        '''

        # identify time map with minimum nhours
        dict_nhours_to_tm = {tm[1]: self.dict_nd_tm_id[nd]
                             for nd, tm in self._dict_nd_tm.items()}
        tm_nhoursmin = dict_nhours_to_tm[min(dict_nhours_to_tm)]

        # identify time map with minimum freq (for joining on hy)

        dict_freq_to_tm = {tm[0]: self.dict_nd_tm_id[nd]
                           for nd, tm in self._dict_nd_tm.items()}
        tm_freqmin = dict_freq_to_tm[min(dict_freq_to_tm)]

        df = self._tm_objs[tm_freqmin].df_time_map.set_index('hy')[[]]

        for tm_id, tm_obj in self._tm_objs.items():
            df = df.join(tm_obj.df_time_map.set_index('hy').sy.rename(tm_id))

        df = df.fillna(method='ffill', axis=0)

        df = df.assign(sy_min=df[tm_nhoursmin]).set_index('sy_min')

        # if the timemap tm_nhoursmin has nhours > freq there are duplicates
        df = df.drop_duplicates().reset_index()

        # stack to make the tm_ids a columns

        self.df_sy_min_all = (df.set_index('sy_min').stack().reset_index()
                                .rename(columns={'level_1': 'tm_id', 0: 'sy'}))

    def _soy_map_hydro_bcs(self):
        ''' Map hydro boundary conditions (which refer to the beginning
            of the month) to time slots
        '''

        if not self.df_plant_month is None:
            df = self.df_def_month[['mt_id', 'month_min_hoy']]
            df = df.set_index('mt_id')
            self.df_plant_month = self.df_plant_month.join(df,
                                                           on=df.index.name)
            self.df_plant_month['tm_id'] = (
                    self.df_plant_month.pp_id
                        .replace(self.mps.dict_plant_2_node_id)
                        .replace(self.dict_nd_tm_id))
            df = self.df_hoy_soy.rename(columns={'hy': 'month_min_hoy'})
            df = df.set_index(['month_min_hoy', 'tm_id'])
            self.df_plant_month.month_min_hoy = (
                self.df_plant_month.month_min_hoy.astype(float))
            self.df_plant_month = self.df_plant_month.join(df,
                                                           on=df.index.names)

            sy_null = self.df_plant_month.sy.isnull()
            self.df_plant_month = self.df_plant_month.loc[-sy_null]


    def map_to_time_res(self):
        '''
        Generates a map between hours-of-the-year and time slots-of-the-year
        based on the fixed time resolution ``self.nhours``, using the class
        ``timemap.TimeMap``. Then maps time series input data from hours
        to slots.
        Also generates dictionaries which contain all slot ids for each
        week/month and vice versa.

        Raises
        ------
        ValueError
            If multiple tm_ids correspond to the pf_ids in any
            selected profile tables. This could be fixed by splitting up
            the profile table, but we rather have this avoided on the
            input data side.

        '''


        if getattr(self, 'df_tm_soy', None) is not None:  # is none by default, so hasattr doesn't work!z
            self._init_time_map_input()
        else:
            self._init_time_map()

        if (isinstance(self.df_node_connect, pd.DataFrame)
            and not self.df_node_connect.empty):
            self._init_time_map_connect()

        self.adjust_cost_time()
        self._soy_map_hydro_bcs()


        # Map profiles and bc to soy
        for itb, idx in [
                         ('dmnd', ['dmnd_pf_id', 'sy']),
                         ('inflow', ['pp_id', 'ca_id', 'sy']),
                         ('supply', ['supply_pf_id', 'sy']),
                         ('chp', ['nd_id', 'ca_id', 'sy']),
                         ('pricesll', ['price_pf_id', 'sy']),
                         ('pricebuy', ['price_pf_id', 'sy']),
                        ]:

            name_df = 'df_prof' + itb
            logger.info('Averaging {}; nhours={}.'.format(name_df,
                                                          self.nhours))

            if hasattr(self, name_df) and getattr(self, name_df) is not None:

                df_tbsoy = getattr(self, name_df)

                setattr(self, 'df_prof' + itb + '_soy',
                        self.map_profile_to_time_resolution(df=df_tbsoy,
                                                            idx=idx, itb=itb))

        self._get_maximum_demand()


    def _add_tm_columns(self, df):
        '''
        Adds a ``tm_id`` column to ``df`` based on suitable other indices.

        Depending on availability, the profile (``pf_id``), plant(``pp_id``)
        or node (``nd_id``) column is used to obtain a timemap (``tm_id``)
        column.

        Parameters
        ----------
        df: DataFrame
            Input table

        Returns
        -------
        DataFrame
            table with additional ``tm_id`` column,

        Raises
        ------
        IndexError
            When ``df`` has neither a ``nd_id`` column nor a ``pp_id`` column
            nor a ``pf_id`` column.
        '''

        cols = df.columns.tolist()

        if not any(col_slct in cols
                   for col_slct in ['pp_id', 'nd_id']):


            list_pf_col = [c for c in cols if 'pf_id' in c]
            if list_pf_col:
                pf_id = df[list_pf_col[0]]
                df = self.translate_pf_id(df.assign(pf_id=pf_id))
            else:
                raise IndexError('_add_tm_columns: '
                                 'no nd_id, pf_id, or pp_id column '
                                 'in table with columns%s.'%cols)

        if 'pp_id' in df.columns:

            dct_p2tm = self.dict_pp_tm_id
            df['tm_id'] = df.pp_id.replace(dct_p2tm)

        elif 'nd_id' in df.columns:

            dct_n2tm = self.dict_nd_tm_id
            df['tm_id'] = df.nd_id.replace(dct_n2tm)

        return df[cols + ['tm_id']]

    import wrapt
#
#    @wrapt.decorator
#    def skip_if_df_none(f, self, **kwargs):
#        print(kwargs)
#        return f(**kwargs)


    def map_profile_to_time_resolution(self, df, idx, itb=None):
        '''
        Maps a single profile table to the selected nodal time resolution.

        If the input table ``df`` doesn't contain a timemap column tm_id,
        it is added based on the ``nd_id`` or ``pp_id`` columns using
        the method :func:`_add_tm_columns`.

        This instance method ca also be used externally to map other
        input data to the temporal structure of the model (e.g. if the
        demand or supply profiles are changed between model runs). Note that
        in this case the ``tm_id`` column must already be present in the ``df``
        (or the indices mapped to the ids of the original model). Otherwise
        the ``tm_id`` will not be found.

        Parameters
        ----------
        df : DataFrame
            input profile table. Must have columns ``hy`` as well as one of
            ``{tm_id, pp_id, nd_id, pf_id}``
        idx: list of str
            index columns of the new tables
        itb: str
            table name, only used in error message

        Returns
        -------
        DataFrame
            Mapped table

        Raises
        ------
        IndexError
            When a given ``pf_id`` is used by multiple ``tm_id``. E.g. if
            a price profile is used by different nodes with different
            time resolutions. Could be implemented... not done for now.

        '''

        df = df.copy()

        val = ['value']

        if df is None or df.empty:
            # empty table
            return pd.DataFrame(columns=idx + val)

        if not 'tm_id' in df.columns:
            df = self._add_tm_columns(df)

        # are there single nodes pf_ids which have more than one tm_id?
        if 'pf_id' in df.columns:
            df_pf = df[['pf_id', 'tm_id']].drop_duplicates()
            tm_count = df_pf.groupby('pf_id').tm_id.count()
            if tm_count.max() > 1:
                raise IndexError('map_to_time_res: Multiple tm_ids '
                                 'found for pf_ids of %s profiles.'%itb)

        ind = ['hy', 'tm_id']
        df[ind] = df[ind].astype(float)
        df = df.join(self.df_hoy_soy.set_index(ind), on=ind)
        df = df.pivot_table(values=val, index=idx,
                            aggfunc=np.mean).reset_index()

        if df.empty:
            df = pd.DataFrame(columns=idx + val)

        return df

    def adjust_cost_time(self):
        '''
        Scale fixed costs for incomplete years.

        If the model year doesn't have 8760 hours, fixed costs are scaled
        to keep levelized costs the same.

        This is relevant if the tm_filt ModelBase parameter is used to
        work with a simplified model version.
        '''

        tm_weight = self.df_tm_soy.groupby('tm_id').weight.sum() / 8760
        tm_weight = tm_weight.to_dict()

        lstfc = [c for c in self.df_plant_encar.columns if c.startswith('fc_')]

        scale_fc = lambda x: x[lstfc] * tm_weight[self.dict_pp_tm_id[x.pp_id]]

        self.df_plant_encar[lstfc] = \
            self.df_plant_encar[lstfc + ['pp_id']].apply(scale_fc, axis=1)

    def _get_maximum_demand(self):
        '''
        Calculation of maximum demand (in MW) with adjusted time resolution.

        Note: Database table def_node is updated in the
        :func:`io.write_runtime_tables` method.

        '''

        if (hasattr(self, 'df_profdmnd_soy')
            and not self.df_profdmnd_soy.empty):
            df = self.df_profdmnd_soy
            df_dmd_params = df.pivot_table(values=['value'],
                                           index='dmnd_pf_id',
                                           aggfunc=[max]).reset_index()
            df_dmd_params.columns = ['pf_id', 'dmnd_max']
            df_dmd_params = self.translate_pf_id(df_dmd_params)
            df_dmd_params = df_dmd_params.set_index('nd_id')[['dmnd_max']]

            self.df_def_node.drop([c for c in self.df_def_node.columns
                                   if c in df_dmd_params.columns],
                                  axis=1, inplace=True)

            self.df_def_node = self.df_def_node.join(df_dmd_params, on='nd_id')

#    def switch_soln_file(self, isolnfile):
#        '''
#        Warmstart and solutionfiles are alternated between two to
#        allow to use the last solution as starting values but avoid
#        excessive disk space use (which would be the case if we kept all
#        solution files).
#
#        Parameters:
#        isolnfile -- binary index of last solution file
#        '''
#        isolnfile = 1 - isolnfile
#        solnfile = os.path.join(TEMP_DIR,
#                                ('manual_soln_file_{uc}_{i}.cplex.sol'
#                                 .format(uc=get_random_suffix(),
#                                         i=str(isolnfile))))
#        return solnfile, isolnfile

    def init_solver(self):
        '''
        Create pyomo Solverfactory instance and adjust parameters.

        '''
        self.dual = po.Suffix(direction=po.Suffix.IMPORT)

        if sys.platform == 'win32':
            self.solver = SolverFactory("cplex")
        elif sys.platform in ['linux2', 'linux']:
            exec_str = ('/opt/ibm/ILOG/CPLEX_Studio1271/cplex/bin/'
                        +'x86-64_linux/cplex')
            self.solver = SolverFactory("cplex", executable=exec_str)
        elif sys.platform == 'darwin':
            exec_str = ('/Applications/CPLEX_Studio128/cplex/bin/'
                        'x86-64_osx/cplex')
            self.solver = SolverFactory("cplex", executable=exec_str)

        if self.nthreads:
            self.solver.set_options('threads=' + str(self.nthreads))

#        fn = 'manual_log_file_{uc}.cplex.sol'.format(uc=get_random_suffix())
#        self.logfile = os.path.join(TEMP_DIR, fn)
#        self.solver._problem_files = (os.path.join(TEMP_DIR, 'pyomo.lp'),)

        # init of solutionfile
        self.isolnfile = 0
#        self.solutionfile, self.isolnfile = self.switch_soln_file(1)
#        self.warmstartfile = None

    def check_valid_indices(self, index):
        '''
        Checks index sets for validity.

        Used in parameter and variable definitions.

        Parameters
        ----------
        index: tuple
            tuple of of pyomo sets

        '''

        is_empty = [pi.name for pi in index if (not pi is None) and not pi]
        is_none = [pi is None for pi in index]
        if any(is_empty) + any(is_none):
            logger.warning(('... failed: set(s) {} is/are '
                            + 'empty or None.').format(str(is_empty)))
            return False
        else:
            return True


    def delete_component(self, comp_name):
        '''
        Drop a component of the pyomo model.

        A single component object is associated with various index objects.
        Because of this, some looping over the vars is required
        to catch 'em all.

        Parameters
        ----------
        comp_name (str): base name of the model component (variable, etc)

        '''

        list_del = [vr for vr in vars(self)
                    if comp_name == vr
                    or vr.startswith(comp_name + '_index')
                    or vr.startswith(comp_name + '_domain')]

        if list_del:
            list_del_str = ', '.join(list_del)
            logger.info('Deleting model components ({}).'.format(list_del_str))

            for kk in list_del:
                self.del_component(kk)


    @contextlib.contextmanager
    def temp_files(self):

        if not self.keepfiles:  # default temporary files
            tmp_dir, logf, warmf, solnf = (None,) * 4

        else:
            if not self.tempdir:
                self.tempdir = os.getcwd()

            tmp_dir = pyutilib.services.TempfileManager.create_tempdir(
                                    prefix='grimsel_temp', dir=self.tempdir)
            logf = os.path.join(tmp_dir, 'logfile.cplex.log')
            warmf = os.path.join(tmp_dir, 'warmstart.cplex.mst')
            solnf = os.path.join(tmp_dir, 'solution.cplex.sol')

        yield tmp_dir, logf, warmf, solnf

        if self.keepfiles:
            pyutilib.services.TempfileManager.clear_tempfiles()


    def run(self, warmstart=False, tmp_dir=None, logf=None, warmf=None, solnf=None):
        '''
        Run the model. Then switch solution/warmstartfile.

        Unless skip_runs is True. Then just create a pro-forma results object.

        Args:
            warmstart (bool): passed to the Solver solve call
        '''

        if self.skip_runs:
            class Result: pass # ad-hoc class mimicking the results object
            self.results = Result()
            self.results.Solver = [{'Termination condition':
                                    'Skipped due to skip_runs=True.'}]
        else:

            slv_kw = dict(tee=self.verbose_solver, keepfiles=self.keepfiles,
                          warmstart=warmstart,
                          symbolic_solver_labels=self.symbolic_solver_labels,
#                          logfile=logf,
#                          solnfile=solnf,
#                          warmstart_file=warmf,
#                          tempdir=tmp_dir
                          )
            self.results = self.solver.solve(self, **slv_kw)
#            self.warmstartfile = self.solutionfile
#            sf, isf = self.switch_soln_file(self.isolnfile)
#            self.solutionfile, self.isolnfile = [sf, isf]

            self._get_objective_value()


    def _get_objective_value(self):
        '''
        Makes the objective value a :class:`ModelBase` instance attribute.

        This assumes that among the objects defined by `list_obj_name`
        only one actually exists.
        '''

        list_name_obj = ['objective_lin', 'objective_quad', 'objective']


        if self.results.solver.termination_condition.key == 'optimal':
            for name_obj in list_name_obj:
                obj = getattr(self, name_obj, False)
                if obj and isinstance(obj, SimpleObjective) and obj.active:
                    self.objective_value = po.value(obj)
        else:
            self.objective_value = np.nan


    def print_is_fixed(self, variable='cap_pwr_new'):
        '''
        Print all fixed elements of the variable given as an input.
        TODO: This assumes pp as first index: needs to change!
        Keyword arguments:
        variable -- name string of a pyomo variable
        '''
        vv = getattr(self, variable)
        print('*'*15 + ' ' + variable + ': ' + '*'*15)
        for i in vv:
            print('vv:', vv)
            if vv[i].fixed:
                print(str(self.mps.dict_pp[i[0]]) + ': is fixed at '
                      + str(vv[i].value))
        print('*'*(33 + len(variable)) + '\n')


    def fill_peaker_plants(self, demand_factor=1.02, reset_to_zero=False,
                           list_peak=[]):
        '''
        Calculate required capacity of designated peaker plants from
        power capacity and demand profiles.
        This serves to avoid infeasibilities due to insufficient installed
        capacity in case the capacities are fixed.
        Keyword arguments:
        demand_factor -- multiplied with the peak load value, this determines
                         the required dispatchable power capacity
        from_variable -- boolean, if True, use variable 'cap_pwr_tot' instead
                         of default 'cap_pwr_leg'
        reset_to_zero -- boolean, if True: stop after the reset-to-zero stage

        '''

        list_attr = ['cap_pwr_leg']

        # reset to zero: get list of peaker plants ppca indices from
        # df_plant_encar then set both cap_pwr_tot and cap_pwr_leg to zero
        if not list_peak:
            _df = self.df_plant_encar
            _list_peak = _df.loc[_df['pp_id'].isin(self.setlst['peak']),
                                  ['pp_id', 'ca_id']].drop_duplicates()
            _list_peak = _list_peak.apply(tuple, axis=1).tolist()
        else:
            _list_peak = list_peak

        for iattr in list_attr:
            for kk in list_peak:
                getattr(self, iattr)[kk] = 0

        if (self.setlst['peak'] or list_peak) and not reset_to_zero:

            slct_attr = 'cap_pwr_leg'

            # Ids all plants considered dispatchable, not peaker plants.
            slct_pp = [pp for pp in
                       self.setlst['pp']
                       + self.setlst['st']
                       + self.setlst['hyrs']
                       if not pp in self.setlst['peak']]

            list_cap = [tuple(list(cc)
                              + [self.mps.dict_plant_2_node_id[cc[0]]]
                              + [getattr(self, slct_attr)[cc].value])
                        for cc in getattr(self, slct_attr) if cc[0] in slct_pp]

            # generate df with all capacities of dispatchable plants
            _df = pd.DataFrame(list_cap, columns=['pp_id', 'ca_id',
                                                  'nd_id', 'cap_pwr'])

            df_dmd_max = self.df_def_node.set_index('nd_id')['dmnd_max']
            df_dmd_max *= demand_factor

            df_cap_tot = pd.DataFrame(_df.pivot_table(values=['cap_pwr'],
                                          index='nd_id', aggfunc=sum),
                                          columns=['cap_pwr'],
                                          index=df_dmd_max.index).fillna(0)
            df_cap_tot = df_cap_tot['cap_pwr']

            dict_nd_pp = {nd: pp for pp, nd in
                          self.mps.dict_plant_2_node_id.items()
                          if pp in list(zip(*_list_peak))[0]}

            df_cap_peak = df_dmd_max - df_cap_tot
            df_cap_peak = df_cap_peak.apply(lambda x: max(0., x))
            df_cap_peak = df_cap_peak.reset_index()
            df_cap_peak['pp_id'] = df_cap_peak['nd_id']
            list_nd = list(dict_nd_pp.keys())
            df_cap_peak = df_cap_peak.loc[df_cap_peak.nd_id.isin(list_nd)]
            df_cap_peak['pp_id'] = df_cap_peak['pp_id'].replace(dict_nd_pp)
            df_cap_peak['ca_id'] = 0

            df_cap_peak = df_cap_peak.set_index(['pp_id', 'ca_id'])
            dict_peak_1 = df_cap_peak[0].to_dict()

            for iattr in list_attr:
                for kk, vv in dict_peak_1.items():
                    getattr(self, iattr)[kk] = vv

    def activation(self, bool_act=False, constraint_list=False,
                   subset=False, verbose=False):
        ''' Changes activation of a list of constraints to bool_act '''

        if subset:
            if type(subset) is not dict:
                _subset = {c: subset for c in constraint_list}
            else:
                _subset = subset
        else:
            _subset = {c: [ii for ii in getattr(self, c)]
                           for c in constraint_list}

        for iconst in constraint_list:
            obj_constr = getattr(self, iconst)

            for ii in _subset[iconst]:
                if bool_act:
                    getattr(self, iconst)[ii].activate()
                else:
                    getattr(self, iconst)[ii].deactivate()

            if verbose:
                if type(verbose) == bool:
                    verbose = len(obj_constr)
                print(obj_constr)
                for ikk, kk in enumerate(obj_constr):
                    if ikk <= verbose :
                        print('{}: {}; is active: {}'.format(kk, obj_constr[kk],
                                                            obj_constr[kk].active))
                print('...\n' if verbose < len(obj_constr) else 'end\n')


    def set_variable_const(self, value=0, variable_list=False, verbose=False):

        for varname in variable_list:
            obj_var = getattr(self, varname)
            keys_var = [c for c in obj_var]
            dict_new = {kk: value for kk in keys_var}
            obj_var.set_values(dict_new)

            if verbose:
                if type(verbose) == bool:
                    verbose = len(obj_var)
                print(varname)
                for ikk, kk in enumerate(obj_var):
                    if ikk <= verbose:
                        print('{}: {}; is fixed: {}'.format(kk, obj_var[kk].value,
                                                            obj_var[kk].fixed))
                print('...\n' if verbose < len(obj_var) else 'end\n')

    def set_variable_fixed(self, bool_fix=True, variable_list=False,
                           subset=False, exclude=False, verbose=False):


        for varname in variable_list:
            obj_var = getattr(self, varname)


            _exclude = exclude if exclude else []
            _subset = [c for c in obj_var if not c in _exclude] if not subset else subset



            for ii in _subset:
                if bool_fix:
                    obj_var[ii].fix()
                else:
                    obj_var[ii].unfix()

            if verbose:
                if type(verbose) == bool:
                    verbose = len(obj_var)
                print(varname)
                for ikk, kk in enumerate(obj_var):
                    if ikk <= verbose and kk in _subset:
                        print('{}: {}; is fixed: {}'.format(kk, obj_var[kk].value,
                                                            obj_var[kk].fixed))
                print('...\n' if verbose < len(obj_var) else 'end\n')



#    def scale_nodes(self, nodes, comp_slct=None):
#        '''
#        Scale relevant components with a node-specific factor.
#
#        Parameters
#        ----------
#        nodes: dict
#            node: factor dictionary with types ``{str: numeric}``
#        comp_slct: list of str
#            Names of components whose values are to be scaled. Subset of
#            ``['dmnd', 'cap_pwr_leg', 'erg_inp', 'erg_chp', 'cap_trme_leg',
#               'cap_trmi_leg']``
#
#        '''
#
#        if not comp_slct:
#            comp_slct = ['dmnd', 'cap_pwr_leg', 'erg_inp', 'erg_chp',
#                         'cap_trme_leg', 'cap_trmi_leg']
#
#        for nd, scale in nodes.items():
#
#            nd_id = self.mps.dict_nd_id[nd]
#
#            list_pp_id = self.df_def_plant.loc[self.df_def_plant.nd_id
#                                               == nd_id].pp_id.tolist()
#
#            mk_ndca = self.df_node_encar.nd_id == nd_id
#            list_ndca_id = self.df_node_encar.loc[mk_ndca, ['nd_id', 'ca_id']]
#            list_ndca_id = list_ndca_id.apply(tuple, axis=1).tolist()
#
#            mk_ppca = self.df_plant_encar.pp_id.isin(list_pp_id)
#            list_ppca_id = self.df_plant_encar.loc[mk_ppca, ['pp_id', 'ca_id']]
#            list_ppca_id = list_ppca_id.apply(tuple, axis=1).tolist()
#
#            if 'dmnd' in comp_slct:
#                list_pf_id_dmnd = [self.dict_dmnd_pf[ndca]
#                                   for ndca in list_ndca_id]
#                for key in self.dmnd.sparse_keys():
#                    if key[-1] in list_pf_id_dmnd:
#                        self.dmnd[key] *= scale
#
#            if 'cap_pwr_leg' in comp_slct:
#                for key in self.cap_pwr_leg.sparse_keys():
#                    if key in list_ppca_id:
#                        self.cap_pwr_leg[key] *= scale
#
#            if 'erg_inp' in comp_slct:
#                for key in self.erg_inp.sparse_keys():
#                    if key[0] == nd_id and self.erg_inp[key].value:
#                        self.erg_inp[key] *=scale
#
#            if 'erg_chp' in comp_slct:
#                for key in self.erg_chp.sparse_keys():
#                    if key in list_ppca_id and self.erg_chp[key].value:
#                        self.erg_chp[key] *=scale
#
#            if 'cap_trme_leg' in comp_slct:
#                for key in self.cap_trme_leg.sparse_keys():
#                    if nd_id in key[1:3]:
#                        self.cap_trme_leg[key] *= scale
#
#            if 'cap_trmi_leg' in comp_slct:
#                for key in self.cap_trmi_leg.sparse_keys():
#                    if nd_id in key[1:3]:
#                        self.cap_trmi_leg[key] *= scale
#










#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import grimsel

logger = grimsel._get_logger(__name__)


logger.info(f'grimsel_config file {__file__}')

try:
    import config_local as conf_local
    PATH_CSV = conf_local.PATH_CSV
except Exception as e:
    print(e)

    PATH_CSV = os.path.join(grimsel.__path__[0], 'input_data')
    print('Using default csv path %s'%PATH_CSV)

try:
    import config_local as conf_local
    BASE_DIR = conf_local.BASE_DIR
except Exception as e:
    print(e)
    logger.warning('Could not import BASE_DIR path.')

try:
    import config_local as conf_local
    DATABASE = conf_local.DATABASE
    SCHEMA = conf_local.SCHEMA

    PSQL_USER = conf_local.PSQL_USER
    PSQL_PASSWORD = conf_local.PSQL_PASSWORD
    PSQL_HOST = conf_local.PSQL_HOST
    PSQL_PORT = conf_local.PSQL_PORT

except Exception as e:
    print(e)
    logger.error('''
Could not read PSQL configuration parameters from
config_local.
Please set configuration parameters in
a new project-specific config_local.py file, e.g.

import os
DATABASE = \'database_name\'
SCHEMA = 'model_input_schema_name'

PSQL_USER = 'user'
PSQL_PASSWORD = 'password'
PSQL_PORT = 5432
PSQL_HOST = 'localhost'
'''
)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 14 07:41:11 2018

@author: mcsoini
"""

import pandas as pd
import itertools as it
import matplotlib as mpl
import matplotlib.pyplot as plt
import abc

from grimsel import _get_logger

logger = _get_logger(__name__)



class AutoComplete(abc.ABC):
    '''
    Base class for the autocompletion of model input DataFrames.
    '''

    new_cols = {}

    def __init__(self, m):

        format_list = [self.df_name, type(self).__name__]
        logger.info('Autocompletion '
                    '{} in {}'.format(*format_list))

        self.m = m

        flag_feas, lst_mss_df = self._check_feasible()

        if flag_feas:
            # Control attributes need to be assigned in child
            # classes, otherwise they are None, which won't work.
            if not hasattr(self, '_df'):
                self._df = None # DataFrame to be completed
            if not hasattr(self, '_add_col'):
                self._add_col = None # Reference column(s)

            self.df_add = None
            self.lst_add = []

            self.get_row_list()
            self.filter_rows()

            if self.lst_add:

                self.generate_df_add()
                self.reset_index()
                self.complement_columns()
                self._set_pf_id_nan()
                self.concatenate()

                logger.info('done. Added: '
                             '{}'.format(', '.join(map(str, self.lst_add))))
            else:
                logger.info('nothing added.')
        else:
            logger.warning('infeasible. Missing model DataFrames: '
                           '{}'.format(', '.join(lst_mss_df)))



    def _check_feasible(self):
        '''
        Return False if DataFrame requirements are not met.

        Some child classes have specific DataFrame requirements. These are
        defined by the lst_req_df instance attribute, if applicable. This
        method checks whether the required tables are available as model
        attributes and returns True or False accordingly.

        Returns
        =======

        flag_feasible : bool
        lst_mss_df : list of missing DataFrames
        '''

        flag_feas = True
        lst_mss_df = None
        if hasattr(self, 'lst_req_df'):
            lst_mss_df = [df for df in self.lst_req_df
                          if (not hasattr(self.m, df)
                              or getattr(self.m, df) is None)]
            if lst_mss_df:
                # set flag to False if any of the required
                # dataframes are missing
                flag_feas = False

        return flag_feas, lst_mss_df

    @abc.abstractmethod
    def get_row_list(self):
        ''' List of rows to potentially be added. Implemented in children. '''

        pass

    def filter_rows_existing(self):
        ''' Remove rows already included in _df. By names in _add_col. '''

        # filter for existing rows
        # everything based on list of tuples; converted accordingly if necessary
        ttt = lambda x: tuple(x)
        slct_col = ([self._add_col]
                    if not type(self._add_col) is list
                    else self._add_col)
        self.lst_add = [pc for pc in self.lst_add
                        if not tuple([pc]
                                     if not type(pc) is list
                                     else pc)
                        in self._df[slct_col].apply(ttt, axis=1).tolist()]

    def filter_rows(self):
        ''' Overridden by child classes if other filters need to be applied '''

        self.filter_rows_existing()

    def generate_df_add(self):
        '''
        Generate full DataFrame to be appended to _df.
        '''

        # some children assign df_add before this (e.g. for pp name generation)
        if self.df_add is None:

            cols = (self._add_col
                    if type(self._add_col) is list
                    else [self._add_col])
            self.df_add = pd.DataFrame(self.lst_add, columns=cols)

    def _set_pf_id_nan(self):
        '''
        Set the values of profile ids in all values to None.

        ``pf_id`` columns must be None if not applicable; the value zero
        if assigned to a specific profile.
        '''

        self.df_add[[c for c in self.df_add.columns if 'pf_id' in c]] = None


    def reset_index(self):
        ''' Implemented in children. Calls _reset_index if required. '''
        pass

    def _reset_index(self):
        '''
        Reset index so it starts from 0, then shift to maximum of _df plus one.

        This is called from the child classes, since not relevant for all.
        E.g. not for non-def_ dataframes, which don't have a
        '''

        id_col = self._add_col + '_id'
        self.df_add = (self.df_add.reset_index(drop=True).reset_index()
                                  .rename(columns={'index': id_col}))
        self.df_add[id_col] += self._df[id_col].max() + 1

    def add_color_col(self, cmap=None):

        _cmap = plt.cm.get_cmap(cmap)

        self.df_add['color'] = ''

        # reset index so we can iterate using loc
        self.df_add = self.df_add.reset_index(drop=True)

        for irow in range(self.df_add.iloc[:, 0].size):
            if cmap is None:
                color = '#ffffff'
            else:
                color = mpl.colors.to_hex(_cmap(irow))

            self.df_add.loc[irow, 'color'] = color


    def add_zero_cols(self):
        ''' Check which columns are missing in df_add and fill with zeros. '''

        list_cols = [c for c in self._df.columns
                     if not c in self.df_add.columns]

        for icol in list_cols:
            self.df_add[icol] = 0


    def concatenate(self):
        '''
        Concatenate existing and new rows.

        Fills the values of the new columns with default values as specified
        in the ``new_cols`` class attribute dictionary.

        '''

        add_cols = set(c for c in self.df_add.columns
                       if c in self._df.columns) | set(self.new_cols)

        add_cols = list(add_cols)

        df_new = pd.concat([self._df, self.df_add[add_cols]], sort=False)

        for col, def_val in self.new_cols.items():
            df_new[col] = df_new[col].fillna(def_val)

        setattr(self.m, self.df_name, df_new)
        self._df = getattr(self.m, self.df_name)

    def complement_columns(self):
        ''' Implemented in child classes if additional columns are required '''
        pass

    def _get_dmnd_list(self, name_type):
        '''
        Returns a list of the relevant fuel or pp_type names.

        The returned list depends on the ``autocomplete_curtailment``
        bool value.

        Args:
            name_type (str): one of ('fuel', 'plant')

        Returns:
            list. The selected list of demand-like plants or fuels.
        '''

        if name_type == 'fuel':
            flx, dmd = 'dmnd_flex', 'dmnd'
        elif name_type == 'plant':
            flx, dmd = 'DMND_FLEX', 'DMND'

        return ([flx, dmd] if self.autocomplete_curtailment else [dmd])



class AutoCompletePpType(AutoComplete):

    df_name = 'df_def_pp_type'

    def __init__(self, m, autocomplete_curtailment):

        self._df = getattr(m, self.df_name)
        self._add_col = 'pt'

        self.autocomplete_curtailment = autocomplete_curtailment

        super().__init__(m)

    def get_row_list(self):
        ''' Static pt names plus one for each energy carrier in def_encar. '''

        self.lst_add = ['TRNS'] + self._get_dmnd_list('plant')
        self.lst_add += ['CONS_' + ca for ca in self.m.df_def_encar.ca]


    def reset_index(self):
        ''' Calls parent _reset_index method. '''

        super()._reset_index()

    def complement_columns(self):
        ''' Add pp_broad_cat and color. '''

        self.df_add['pp_broad_cat'] = self.df_add['pt']
        self.add_color_col('Set1')

class AutoCompleteFuel(AutoComplete):

    df_name = 'df_def_fuel'
    def __init__(self, m):

        self._df = getattr(m, self.df_name)
        self._add_col = 'fl'

        # define column is_ca based on entries of df_def_encar
        self._df['is_ca'] = 0
        if 'fl_id' in m.df_def_encar.columns:
            mask_ca = self._df.fl_id.isin(m.df_def_encar.fl_id)
            self._df.loc[mask_ca, 'is_ca'] = 1

        super().__init__(m)

    def reset_index(self):
        ''' Calls parent class _reset_index '''

        self._reset_index()

    def complement_columns(self):

        self.add_color_col('Set3')
        self.add_zero_cols()

class AutoCompleteTrns():
    ''' Mixin class holding methods relevant for inter-node transmission. '''

    def filter_multi_node(self):
        ''' Delete all entries if there is only one node.'''

        self.lst_add = (self.lst_add
                        if self.m.df_def_node.nd_id.count() > 1
                        else [])

class AutoCompleteFuelTrns(AutoCompleteFuel, AutoCompleteTrns):

    def __init__(self, m):

        super().__init__(m)

    def get_row_list(self):
        ''' Static fl names. '''

        self.lst_add = ['exchange']

    def filter_rows(self):

        super().filter_multi_node()
        super().filter_rows_existing()



class AutoCompleteFuelConsumed(AutoCompleteFuel):

    def __init__(self, m):

        super().__init__(m)

    def get_row_list(self):
        ''' fl names generated from ca names in df_def_encar. '''

        self.lst_add = self.m.df_def_encar['ca'].tolist()
        self.lst_add = ['consumed_' + fl.lower() for fl in self.lst_add
                        if not fl in self._df[self._add_col]]

class AutoCompletePlant(AutoComplete):

    df_name = 'df_def_plant'
    def __init__(self, m):

        self._df = getattr(m, self.df_name)
        self._add_col = 'pp'

        super().__init__(m)

    def reset_index(self):
        ''' Calls parent class _reset_index '''

        self._reset_index()

    def add_id_cols(self, add_names=False):
        '''
        Add id columns based on name columns.

        Argument:
        add_names -- bool; add name columns based on id columns instead.
        '''

        for name, df_def in [('nd', 'df_def_node'),
                             ('fl', 'df_def_fuel'),
                             ('pt', 'df_def_pp_type')]:

            _name = name
            _id = name + '_id'

            if add_names:
                # reverse
                _name = name + '_id'
                _id = name

            if ((not _id in self.df_add.columns)
                and (_name in self.df_add.columns)):

                mp = getattr(self.m, df_def).set_index(_name)[_id]
                self.df_add = self.df_add.join(mp, on=_name)

class AutoCompletePlantTrns(AutoCompletePlant, AutoCompleteTrns):

    def __init__(self, m):
        # list of required model dataframes

        self.lst_req_df = ('df_node_connect',)

        super().__init__(m)

    def get_row_list(self):
        ''' pp names from reshaped df_node_connect table. '''

        mask_trns = ((self.m.df_node_connect.cap_trmi_leg != 0) |
                     (self.m.df_node_connect.cap_trme_leg != 0))
        self.df_add = (self.m.df_node_connect
                           .loc[mask_trns, ['nd_id', 'nd_2_id']]
                           .drop_duplicates()
                           .stack()
                           .reset_index()
                           .rename(columns={'level_1': 'imex', 0: 'nd_id'})
                           .join(self.m.df_def_node.set_index('nd_id')['nd'],
                                 on='nd_id')[['imex', 'nd_id', 'nd']]
                           .drop_duplicates())

        self.df_add['pt'] = 'TRNS'
        self.df_add['fl'] = 'exchange'
        self.df_add['pp'] = self.df_add['nd'] + '_' + self.df_add['pt']

        self.lst_add = self.df_add['pp'].tolist()

    def filter_rows(self):

        super().filter_multi_node()
        super().filter_rows_existing()

    def complement_columns(self):

        self.add_color_col('Accent')
        self.add_id_cols()
        self.add_zero_cols()

        self.df_add['set_def_tr'] = 1


class AutoCompleteFuelDmnd(AutoCompleteFuel):

    def __init__(self, m, autocomplete_curtailment):

        self.autocomplete_curtailment = autocomplete_curtailment

        super().__init__(m)


    def get_row_list(self):
        ''' Static fl names. '''

        self.lst_add = self._get_dmnd_list('fuel')

class AutoCompletePlantDmnd(AutoCompletePlant):

    new_cols = {'set_def_dmd': 0, 'set_def_curt': 0}

    def __init__(self, m, autocomplete_curtailment):

        self.autocomplete_curtailment = autocomplete_curtailment

        super().__init__(m)

    def get_row_list(self):
        ''' Cross product fixed pt names and nodes. '''

        self.lst_add = self.m.df_def_node['nd'].tolist()
        self.lst_add = ['_'.join(pp) for pp in
                        list(it.product(self.lst_add,
                                        self._get_dmnd_list('plant')))]

    def complement_columns(self):
        '''

        '''

        get_nd = lambda x: x.replace('_DMND', '').replace('_FLEX', '')
        self.df_add['nd'] = self.df_add.pp.apply(get_nd)

        get_pt = lambda x: x.pp.replace(x.nd + '_', '')
        self.df_add['pt'] = self.df_add[['pp', 'nd']].apply(get_pt, axis=1)

        self.df_add['fl'] = self.df_add['pt'].apply(str.lower)

        self.add_id_cols()
        self.add_zero_cols()

        self.df_add['set_def_dmd'] = self.df_add['set_def_curt'] = 0
        mask_dmnd = ~self.df_add.pt.str.contains('FLEX')
        self.df_add.loc[mask_dmnd, 'set_def_dmd'] = 1

        mask_flex = self.df_add.pt.str.contains('FLEX')
        self.df_add.loc[mask_flex, 'set_def_curt'] = 1


class AutoCompletePlantCons(AutoCompletePlant):

    def __init__(self, m):

        super().__init__(m)

    def get_row_list(self):
        ''' Plants in df_def_plant which consume energy carriers. '''

        list_cafl = self.m.df_def_encar.fl_id.tolist()
        self.df_add = self._df.loc[self._df.fl_id.isin(list_cafl),
                                   ['nd_id', 'fl_id']].drop_duplicates()
        df_flca = self.m.df_def_encar.set_index('fl_id')['ca']
        self.df_add = self.df_add.join(df_flca, on='fl_id')

        # adding nd column
        self.add_id_cols(add_names=True)

        self.df_add['pt'] = 'CONS_' + self.df_add['ca']
        self.df_add['pp'] = self.df_add.nd + '_' + self.df_add.pt

        self.lst_add = self.df_add.pp.tolist()

    def complement_columns(self):

        self.add_id_cols()
        self.add_zero_cols()


class AutoCompletePpCa(AutoComplete):

    df_name = 'df_plant_encar'
    completion_type = ''

    def __init__(self, m):

        self._df = getattr(m, self.df_name)
        self._add_col = ['pp_id', 'ca_id']

        super().__init__(m)

    def filter_rows(self):
        self.filter_rows_existing()


    def complement_columns(self):

        self.add_zero_cols()


class AutoCompletePpCaFlex(AutoCompletePpCa):

    def __init__(self, m, autocomplete_curtailment):

        self.autocomplete_curtailment = autocomplete_curtailment

        super().__init__(m)

    def get_row_list(self):
        ''' . '''

        if self.autocomplete_curtailment:

            mask_flex = self.m.df_def_plant.pp.str.contains('FLEX')
            list_pp = self.m.df_def_plant.loc[mask_flex, 'pp_id'].tolist()
            list_ca = self.m.df_def_encar['ca_id'].tolist()

            self.lst_add = [list(pc) for pc in it.product(list_pp, list_ca)]
        else:
            self.lst_add = []


if __name__ == '__main__':

    self = AutoCompletePpType(ml.m)
    self = AutoCompleteFuelTrns(ml.m)
    self = AutoCompleteFuelDmnd(ml.m)
    self = AutoCompletePlantTrns(ml.m)
#    self = AutoCompletePlantCons(ml.m)
    self = AutoCompletePpCaFlex(ml.m)

    ml.m.df_def_plant = ml.m.df_def_plant.loc[~ml.m.df_def_plant.pp.str.contains('DMND')]
    ml.m.df_def_plant = ml.m.df_def_plant.drop('set_def_dmnd', axis=1)
    self = AutoCompletePlantDmnd(ml.m, True)

# %%











import os
import numpy as np
import pandas as pd
import wrapt
from glob import glob
import re

import grimsel.auxiliary.sqlutils.aux_sql_func as aql
from grimsel.auxiliary.aux_general import silence_pd_warning
from grimsel import _get_logger

logger = _get_logger(__name__)

rev_dict = lambda dct: {val: key for key, val in dct.items()}


class Maps():
    '''
    Transforms definition tables into dictionaries.

    Uses the input tables with prefix *def_* to define dictionaries between
    ids and names. This includes
    - ``dict_node = {node id (int): node name (str)}``
    - ``dict_plant = {plant id (int): plant name (str)}``
    - ``dict_pp_type = {pp_type id (int): pp_type name (str)}``
    - ``dict_fuel = {fuel id (int): fuel name (str)}``
    - ``dict_profile = {profile id (int): profile name (str)}``


    Constructors:
    - default: PSQL schema and database.
    - ``from_df``: List of DataFrames.


    '''

    list_id_tbs = {'fuel': 'fl',
                   'profile': 'pf',
                   'encar': 'ca',
                   'node': 'nd',
                   'plant': 'pp',
                   'pp_type': 'pt',
                   'run': 'run'}

    list_id_tbs_rev = {val: key for key, val in list_id_tbs.items()}


    def __init__(self, sc, db, dict_tb=None):

        self.sc = sc
        self.db = db

        if not dict_tb:
            self._dict_tb = self._read_tables_sql()
        else:
            self._dict_tb = self._adjust_input_tables(dict_tb)

#
#        if 'run' in self._dict_tb and self._dict_tb['run'] is not None:
#            self.list_id_tbs['run'] = list(c for c in self._dict_tb['run']
#                                           if c.endswith('_vl'))

        self._make_id_dicts()
        self._make_color_dicts()


    def _adjust_input_tables(self, dict_tb):
        '''
        Adjust input table formatting.

        Makes sure that the tables provided through the alternative
        initializer :method:``from_dicts`` are formatted in the same way
        as the ones from the SQL input (:method:`_read_tables_sql`).
        '''

        for name, df in dict_tb.items():

            logger.debug('{}, {}'.format(name, Maps.list_id_tbs[name]))
            name_idx = Maps.list_id_tbs[name] + '_id'


            if name_idx in df.columns:

                df.set_index(name_idx, inplace=True)

        return dict_tb


    @classmethod
    def from_dicts(cls, dict_tb):
        '''
        Alternative constructor accepting input tables in a dictionary.

        Args:
            :type:`dict` like ``{short table name (str): DataFrame}``

        '''

        dict_tb = {key: df.copy() for key, df in dict_tb.items()
                   if key in Maps.list_id_tbs}

        self = cls(None, None, dict_tb)

        return self

    @classmethod
    def from_hdf5(cls, fn):

        with pd.HDFStore(fn) as store:

            keys = store.keys()

            dict_tb = {key.replace('/def_', ''): store.get(key)
                       for key in keys
                       if key.replace('/', '').replace('def_', '')
                       in Maps.list_id_tbs}

        if not dict_tb:
            raise IOError('File %s not found.'%fn)

        self = cls(None, None, dict_tb)

        return self

    @classmethod
    def from_parquet(cls, dirc):

        dict_fn_key = {fn: fn.split(os.sep)[-1]
                             .replace('def_', '').replace('.parq', '')
                       for fn in glob(dirc + '/*.parq')}

        dict_tb = {key: pd.read_parquet(fn)
                   for fn, key in dict_fn_key.items()
                   if key in Maps.list_id_tbs}

        if not dict_tb:
            raise IOError('Parquet directory %s not found.'%dirc)

        self = cls(None, None, dict_tb)

        return self


    def _read_tables_sql(self):
        '''
        Reads all relevant tables from the input PSQL schema.

        Returns:
            :returns: :type:`dict` like ``{short table name (str): DataFrame}``

        '''

        list_tb_sql = aql.get_sql_tables(self.sc, self.db)
        dict_tb = {}

        for inme, iind in Maps.list_id_tbs.items():

            tb_name = 'def_' + inme

            if tb_name in list_tb_sql:
                 tb = aql.read_sql(self.db, self.sc, tb_name)
                 tb = tb.set_index(iind + '_id')
                 dict_tb[inme] = tb

        return dict_tb



    def display_colormap(self):
        all_cols = {'c' + str(i): getattr(self, 'c' + str(i))
                    for i in range(1, 200)
                    if 'c' + str(i) in self.__dict__.keys()}

        df = pd.DataFrame(np.abs(np.sin(np.random.randn(2, len(all_cols)))),
                          columns=all_cols.keys())
        df.plot.area(stacked=True, color=[all_cols[c] for c in df.columns])

    def get_color_dict(self, iind, list_complement=[], color_complement='k'):
        '''
        Returns a color dictionary based on the *color* columns in the input.

        Parameters
        ----------
        iind : str
            index like fl, fl_id, pp, pp_id, etc.
        list_complement : list
            list of entries which are not included in the input table but
            which are required in the color dictionary
        color_complement : str
            color string of any format for the complemented color dictionary
            entries

        Returns
        -------
        dict
            dictionary ``{index: color}``

        bool
            Returns ``False`` if the the corresponding table doesn't exist
            or if it has no *color* column.


        '''

        dict_name = '_color_' + iind
        if hasattr(self, dict_name):
            color_dict = getattr(self, '_color_' + iind)

            color_dict_compl = {key: color_complement
                                for key in list_complement
                                if not key in color_dict}

            color_dict.update(color_dict_compl)

            return color_dict

        else:
            return False

    def _make_color_dicts(self):
        '''

        Note:
            Color maps for plants are generated from the plant type color maps,
            if required.
        '''


        inme = 'fuel'
        df = self._dict_tb[inme]
        for inme, df in self._dict_tb.items():

            iind = Maps.list_id_tbs[inme]

            if 'color' in df.columns:
                setattr(self, '_color_%s'%iind, df.set_index(iind)['color']
                                                 .to_dict())

                setattr(self, '_color_%s_id'%iind, df['color'].to_dict())

        if (hasattr(self, 'dict_plant_2_pp_type')
            and hasattr(self, '_color_pt')
            and not hasattr(self, '_color_pp')):
            # generate pp color map from pt

            self._color_pp_id = {pp_id: self._color_pt_id[pt] for pp_id, pt
                                 in self.dict_plant_2_pp_type.items()}
            self._color_pp = {self.dict_pp[pp_id]: col for pp_id, col
                              in self._color_pp_id.items()}

    def _make_id_dicts(self):
        '''
        '''

        for inme, df in self._dict_tb.items():
            if not inme == 'run':
                idx = Maps.list_id_tbs[inme]

                dict_0 = df[idx].to_dict()
                setattr(self, 'dict_%s'%idx, dict_0)
                setattr(self, 'dict_%s_id'%idx, rev_dict(dict_0))

        self.dict_nd_2 = getattr(self, 'dict_nd', None)
        self.dict_nd_2_id = getattr(self, 'dict_nd_id', None)


        self.dict_plant_2_node_id = self._dict_tb['plant']['nd_id'].to_dict()
        self.dict_plant_2_pp_type_id = self._dict_tb['plant']['pt_id'].to_dict()
        self.dict_plant_2_fuel_id = self._dict_tb['plant']['fl_id'].to_dict()


        def get_name_dict(name, id2):
            df = self.id_to_name(self._dict_tb[name][[id2 + '_id']], [id2])
            return df[id2 + '_id'].to_dict()

        self.dict_plant_2_pp_type = get_name_dict('plant', 'pt')
        self.dict_plant_2_node = get_name_dict('plant', 'nd')



    def _add_pp_maps(self, df, inplace=False):

        cols = ['nd_id', 'fl_id', 'pt_id']
        cols = list(set(cols) - set(df.columns))

        pp_df = self._dict_tb['plant'][cols]

        if not inplace:
            df = df.copy()

        return df.join(pp_df, on='pp_id')

    @silence_pd_warning
    def id_to_name(self, df, name_list=None, inplace=False,
                   keep_cols=True):

        if 'pp_id' in df.columns:
            df = self._add_pp_maps(df, inplace)

        if not name_list:
            name_list = [c.replace('_id', '') for c
                         in df.columns if c.endswith('_id')]

        if not inplace:
            df = df.copy()

        for iid in name_list:
            idict = getattr(self, 'dict_' + iid, None)

            if idict:

                col_name = iid + ('_id' if not keep_cols else '')
                df.loc[:, col_name] = df[iid + '_id'].replace(idict)
            elif iid == 'run':
                df = self.run_id_to_names(df)

        return df

    def run_id_to_names(self, df):

        if 'run' in self._dict_tb:
            ddfrun = self._dict_tb['run']
            ddfrun = ddfrun[[c for c in ddfrun.columns
                             if re.match('.*[_vl|_id]$', c)]]

            df = df.join(ddfrun, on='run_id')

        return df

    @wrapt.decorator
    def param_to_list(f, self, *args, **kwargs):
        if not isinstance(args[0][0], (set, list, tuple)):
            args = (([args[0][0]],),)

        return f(*args[0], **kwargs)

    @param_to_list
    def nd2nd(self, list_nd):
        df = self._dict_tb['node']
        if list_nd:
            return set(df.loc[df.nd.str.contains('|'.join(list_nd))].index.tolist())
        else:
            return {}

    @param_to_list
    def nd2pp(self, list_nd):
        if list_nd:
            df = self._dict_tb['plant']
            return set(df.loc[df.nd_id.isin(self.nd2nd(list_nd))].index.tolist())
        else:
            return {}


    @param_to_list
    def pt2pt(self, list_pt):
        if list_pt:
            df = self._dict_tb['pp_type']
            return set(df.loc[df.pt.str.contains('|'.join(list_pt))].index.tolist())
        else:
            return {}

    @param_to_list
    def pp2pp(self, list_pp):
        if list_pp:
            df = self._dict_tb['plant']
            return set(df.loc[df.pp.str.contains('|'.join(list_pp))].index.tolist())
        else:
            return {}

    @param_to_list
    def fl2fl(self, list_fl):
        if list_fl:
            df = self._dict_tb['fuel']
            return set(df.loc[df.fl.str.contains('|'.join(list_fl))].index.tolist())
        else:
            return {}

    @param_to_list
    def pt2pp(self, list_pt):
        if list_pt:
            df = self._dict_tb['plant']
            return set(df.loc[df.pt_id.isin(self.pt2pt(list_pt))].index.tolist())
        else:
            return {}

    @param_to_list
    def fl2pp(self, list_fl):
        if list_fl:
            df = self._dict_tb['plant']
            return set(df.loc[df.fl_id.isin(self.fl2fl(list_fl))].index.tolist())
        else:
            return {}



if __name__ == '__main__':

    self = Maps('lp_input_levels', 'storage2')

#    ml.m.init_maps()
#
#    self = ml.m.mps



'''
Module doc
'''
import os
from multiprocessing import Lock, Pool, current_process
import numpy as np
import pandas as pd
import itertools
from glob import glob
import time
from importlib import reload
import fastparquet as pq

import grimsel.core.model_base as model_base
import grimsel.core.io as io
import grimsel.core.model_loop_modifier as model_loop_modifier
import grimsel.auxiliary.sqlutils.aux_sql_func as aql
import grimsel.auxiliary.maps as maps
from grimsel import _get_logger

logger = _get_logger(__name__)
logger_parallel = _get_logger(__name__ + '_parallel')


reload(model_base)
reload(io)
reload(model_loop_modifier)

class ModelLoop():
    '''
    Defines the model loop framework. Main attribute is the dataframe
    df_def_run.

     -> variables:
         - dicts
         - def_run
     -> functions:
         - generate run name
         - _print_run_title
         - init loop table
         - reset_run_id
         - generate_loop_row <- from run_id and current values of dictionaries
                                (option for zero row)
    '''

    @property
    def df_def_run(self):
        return self._df_def_run

    @df_def_run.setter
    def df_def_run(self, df_def_run):
        self._df_def_run = df_def_run
        self._df_def_run = self.restore_run_id(self._df_def_run)

    def __init__(self, **kwargs):
        '''
        Keyword arguments:
        nsteps -- list of model loop dimensions and steps; format:
                  (name::str, number_of_steps::int, type_of_steps::function)
        '''

        defaults = {
                    'nsteps': [],
                    'mkwargs': {},
                    'iokwargs': {},
                    'full_setup': True
                    }

        for key, val in defaults.items():
            setattr(self, key, val)
        self.__dict__.update(kwargs)

        self.run_id = None  # set later
        self.__runlevel_state = -1

        self.m = model_base.ModelBase(**self.mkwargs)

        self.iokwargs.update({'model': self.m})
        self.io = io.IO(**self.iokwargs)

        self.init_run_table()
        self.select_run(0)


    @property
    def _runlevel_state(self):
        return self.__runlevel_state

    @_runlevel_state.setter
    def _runlevel_state(self, _runlevel_state):

        assert _runlevel_state == self.__runlevel_state + 1, (
                'Must increment runlevel state by one. Caught attempt to jump '
                f'from {self.__runlevel_state} '
                f'({self._dict_runlevels[self.__runlevel_state]})'
                f' to {_runlevel_state} '
                f'({self._dict_runlevels[_runlevel_state]})')

        self.__runlevel_state = _runlevel_state


    _dict_runlevels = {
            0: 'io.read_model_data',
            1: 'm.init_maps',
            2: 'm.map_to_time_res',
            3: 'io.write_runtime_tables',
            4: 'm.get_setlst',
            5: 'm.define_sets',
            6: 'm.add_parameters',
            7: 'm.define_variables',
            8: 'm.add_all_constraints',
            9: 'm.init_solver',
            10: 'io.init_output_tables'}


    def build_model(self, to_runlevel='full'):
        '''
        Run sequence of methods to read the data and initialize the Pyomo
        components.

        Parameters
        ----------
        to_runlevel : str, one of `['full', 'input_data']`
            `'input_data'`: stop after reading all input data; allows to
            make modifications to the input dataframes
            `'full'`: complete construction of the model; allows to
            make modications to the Pyomo components
        '''

        dict_to_runlevel = {'input_data': 2, 'full': 10}

        assert to_runlevel in dict_to_runlevel, (
                f'Unknown to_runlevel level \'{to_runlevel}\'. '
                f'Expecting one of {list(dict_to_runlevel)}.')

        # filter runlevel dictionary by selected runlevels
        _dict_runlevel_slct = {lvl: meth for lvl, meth
                               in self._dict_runlevels.items()
                               if lvl > self._runlevel_state
                               and lvl <= dict_to_runlevel[to_runlevel]}

        for runlevel, method in _dict_runlevel_slct.items():

            attr, method = method.split('.')

            func = getattr(getattr(self, attr), method)

            logger.info('%' * 60)
            logger.info(f'ModelLoop.build_model: Runlevel {runlevel}: '
                        f'Calling method {method}')
            logger.info('%' * 60)

            func()
            self._runlevel_state = runlevel


    def init_run_table(self):
        '''
        Initializes the ``df_def_run`` table by expanding the ``nsteps`` list.

        Expands the ``nsteps`` parameter to the corresponding DataFrame.
        The resulting attribute ``df_def_run`` contains all relevant
        combinations of the model change indices as defined in ``nsteps``.

        Also initializes the output ``def_run`` table, if required.
        '''

        for istep in self.nsteps:
            assert (isinstance(istep[0], str) and isinstance(istep[1], int)
                    and (len(istep) == 2
                         or istep[2] in (np.linspace, np.arange))), (
                         'Items of nlist must be tuples like '
                         ' (str, int, np.arange or np.linspace)'
                         f' or (str, int); found {istep}')

        _nsteps = [list(ist) + ([0, 1] if ist[-1] == np.linspace else [])
                  for ist in self.nsteps]

        def getfunc(istep): return istep[2] if len(istep) > 2 else np.arange

        list_steps = [list(getfunc(istep)(*istep[3:], istep[1]))
                      for istep in _nsteps]

        list_steps = [list(map(float, lst)) for lst in list_steps]
        full_steps = np.array([tuple(reversed(lst)) for lst in
                      list(itertools.product(*list(reversed(list_steps))))])
        list_index = [list(range(len(i))) for i in list_steps]
        full_index = np.array([tuple(reversed(lst)) for lst in
                      list(itertools.product(*list(reversed(list_index))))])
        full_all = np.concatenate([full_index, full_steps,
                                   float('nan')*full_index], axis=1)
        self.cols_step = [ist[0] for ist in _nsteps]
        self.cols_id = [c + '_id' for c in self.cols_step]
        self.cols_val = [c + '_vl' for c in self.cols_step]
        cols_all = self.cols_id + self.cols_step + self.cols_val

        self.df_def_run = pd.DataFrame(full_all, columns=cols_all)

        if not self.io.resume_loop:
            self.init_loop_table()

    def select_run(self, slct_run_id):
        '''
        Get all relevant indices and parameters for a certain slct_run_id
        '''

        self.run_id = slct_run_id

        lpsrs = self.df_def_run.loc[self.df_def_run.run_id
                                     == slct_run_id].iloc[0]

        self.dct_vl = lpsrs.loc[lpsrs.index.str.contains('_vl')].to_dict()
        self.dct_id = lpsrs.loc[lpsrs.index.str.contains('_id')
                               & ~lpsrs.index.str.contains('run')].to_dict()
        self.dct_step = lpsrs.loc[~lpsrs.index.str.contains('_vl')
                                 & ~lpsrs.index.str.contains('_id')].to_dict()
        self.dct_id = {k: int(v) for (k, v) in self.dct_id.items()}

        self.loop_series = lpsrs


    def init_loop_table(self):

        self.io._init_loop_table(self.cols_id, self.cols_step, self.cols_val)


    def _get_row_df_run(self, tdiff_solve=0, tdiff_write=0, info=''):
        '''
        Generate new row for the def_run table.

        This contains the parameter variation indices as well as information
        on the run (time, objective function, solver status).
        '''

        dtypes = {int: ['run_id'] + list(self.dct_id),
                  float: (['tdiff_solve', 'tdiff_write', 'objective']
                          + list(self.dct_step.keys())),
                  str: ['info'] + list(self.dct_vl)}
        dtypes = {col: dtp  for dtp, cols in dtypes.items() for col in cols}

        vals = [[tdiff_solve, tdiff_write] + [self.run_id] + [info]
                + list(self.dct_id.values())
                + list(self.dct_step.values())
                + list(self.dct_vl.values())]
        cols = (['tdiff_solve', 'tdiff_write', 'run_id', 'info']
                + list(self.dct_id)
                + list(self.dct_step)
                + list(self.dct_vl))

        df_add = pd.DataFrame(vals, columns=cols)

        df_add['objective'] = (self.m.objective_value
                               if hasattr(self.m, 'objective_value')
                               else 0)

        return df_add.astype(dtypes)

    def get_def_run_name(self):

        # if multiprocessing, locked writing to common parquet file has
        # too much overhead for small models. Therefore writing to files
        # by worker + later merge
        if current_process().name == 'MainProcess':
            suffix = ''
            fn = os.path.join(self.io.cl_out, 'def_run%s.parq'%suffix)
        elif current_process().name.startswith('ForkPoolWorker'):
            suffix = '_' + current_process().name
            fn = os.path.join(self.io.cl_out, 'def_run%s.csv'%suffix)
        else:
            raise ValueError('Unexpected current_process name'
                             ' %s'%current_process().name)

        csv_def_run = suffix !=''

        return fn, csv_def_run


    def append_row(self, **kwargs):
        '''
        Generate single-line pandas.DataFrame to be appended to the
        output def_run table. Options:
        - zero_row == True: Row filled with zeros for calibration run
        - zero_row == False: Loop params copied to row
        '''

        df_add = self._get_row_df_run(**kwargs)

        # can't use io method here if we want this to happen when no_output
        if self.io.modwr.output_target == 'psql':
            aql.write_sql(df_add, self.io.sql_connector.db,
                          self.io.cl_out, 'def_run', 'append')
        elif self.io.modwr.output_target == 'hdf5':
            with pd.HDFStore(self.io.cl_out, mode='a') as store:
                store.append('def_run', df_add, data_columns=True,
                             min_itemsize=150 # set string length!
                             )
        elif self.io.modwr.output_target == 'fastparquet':

            fn, csv_def_run = self.get_def_run_name()

            if not csv_def_run:
                pq.write(fn, df_add, append=os.path.isfile(fn))

            else:
                # row-wise appending to parquet is slow for larger amounts of model runs
                # therefore using csv. Merged to parquet in self._merge_df_run_files
                if os.path.isfile(fn):
                    df_add.to_csv(fn, mode='a', header=False, index=False)
                else:
                    df_add.to_csv(fn, header='column_names', index=False)

        else:
            raise ValueError('Unknown output_target '
                             '%s'%self.io.modwr.output_target)


    def _merge_df_run_files(self):
        '''
        Merge all files with name out_dir/def_run_ForkPoolWorker-%d into single
        def_run.
        '''

        list_fn = glob(os.path.join(self.io.cl_out,
                                    'def_run_ForkPoolWorker-[0-9]*.csv'))

        df_def_run = pd.concat(pd.read_csv(fn) for fn in list_fn)
        df_def_run = df_def_run.sort_values('run_id').reset_index(drop=True)

        fn = os.path.join(self.io.cl_out, 'def_run.parq')
        pq.write(fn, df_def_run, append=False)


    def _print_run_title(self, warmstartfile, solutionfile):

        sep = '*' * 60
        run_id_str = 'run_id = %d of %d'%(self.run_id,
                                          self.df_def_run['run_id'].max())
        sw_strs = [str(c[0]) + ' = ' + str(c[1])
                  for c in self.dct_vl.items()]

        logger.info(sep)
        logger.info(run_id_str)
        for strg in sw_strs:
            logger.info(strg)
        logger.info(sep)

        logger.info(run_id_str + ' on ' + current_process().name)

    @staticmethod
    def restore_run_id(df):
        '''
        Reset run_id index after external manipulation of the df_def_run
        '''

        cols_not_run_id = [c for c in df.columns
                           if not c == 'run_id']
        df = df[cols_not_run_id]
        df = df.reset_index(drop=True)
        df = df.reset_index()
        df = df.rename(columns={'index': 'run_id'})

        return df


    def get_list_run_id(self):

        return list(range(self.io.resume_loop,
                          len(self.df_def_run.run_id.tolist())))


    def perform_model_run(self, warmstart=False):
        """
        TODO: This is a mess.

        Calls model_base run methods, io writing methods, and appends to
        def_run. Also takes care of time measurement for reporting in
        the corresponding def_run columns.

        """

        t = time.time()

        with self.m.temp_files() as (tmp_dir, logf, warmf, solnf):

            self._print_run_title(self.m.warmstartfile, self.m.solutionfile)
            self.m.run(warmstart=warmstart, tmp_dir=tmp_dir,
                       logf=logf, warmf=warmf, solnf=solnf)
            tdiff_solve = time.time() - t
            stat = ('Solver: ' + str(self.m.results.Solver[0]['Termination condition']))

            if self.io.replace_runs_if_exist and self.io.resume_loop:

                self.io.delete_run_id(self.run_id, operator='=')

            # append to output tables
            t = time.time()
            self.io.write_run(run_id=self.run_id)
            tdiff_write = time.time() - t

            # append to def_run table
            self.append_row(info=stat,
                            tdiff_solve=tdiff_solve, tdiff_write=tdiff_write)







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan  2 17:26:27 2019

@author: user
"""

var_sy = [
          ('pwr', ('sy', 'pp_id', 'ca_id', 'bool_out')),
          ('dmnd_flex', ('sy', 'nd_id', 'ca_id', 'bool_out')),
          ('pwr_st_ch', ('sy', 'pp_id', 'ca_id', 'bool_out'), 'pwr'),
          ('erg_st', ('sy', 'pp_id', 'ca_id'))
         ]
var_mt = [
    ('erg_mt', ('mt_id', 'pp_id', 'ca_id'))]
var_yr = [
    ('erg_yr', ('pp_id', 'ca_id', 'bool_out')),
    ('erg_fl_yr', ('pp_id', 'nd_id', 'ca_id', 'fl_id')),
    ('pwr_ramp_yr', ('pp_id', 'ca_id')),
    ('vc_fl_pp_yr', ('pp_id', 'ca_id', 'fl_id')),
    ('vc_ramp_yr', ('pp_id', 'ca_id')),
    ('vc_co2_pp_yr', ('pp_id', 'ca_id')),
    ('vc_om_pp_yr', ('pp_id', 'ca_id')),
    ('fc_om_pp_yr', ('pp_id', 'ca_id')),
    ('fc_cp_pp_yr', ('pp_id', 'ca_id')),
    ('fc_dc_pp_yr', ('pp_id', 'ca_id')),
    ('cap_pwr_rem', ('pp_id', 'ca_id')),
    ('cap_pwr_tot', ('pp_id', 'ca_id')),
    ('cap_erg_tot', ('pp_id', 'ca_id')),
    ('cap_pwr_new', ('pp_id', 'ca_id')),
    ('erg_ch_yr', ('pp_id', 'ca_id', 'bool_out'), 'erg_yr')]
var_tr = [
    ('trm', ('sy', 'nd_id', 'nd_2_id', 'ca_id', 'bool_out')),
    ('erg_trm_rv_yr', ('nd_id', 'nd_2_id', 'ca_id', 'bool_out')),
    ('erg_trm_sd_yr', ('nd_id', 'nd_2_id', 'ca_id', 'bool_out'))]
par = [
    ('share_ws_set', ('nd_id',)),
    ('price_co2', ('nd_id',)),
    ('co2_int', ('fl_id',)),
    ('cap_pwr_leg', ('pp_id', 'ca_id')),
    ('cap_avlb', ('pp_id', 'ca_id')),
    ('cap_trme_leg', ('mt_id', 'nd_id', 'nd_2_id', 'ca_id')),
    ('cap_trmi_leg', ('mt_id', 'nd_id', 'nd_2_id', 'ca_id')),
    ('cf_max', ('pp_id', 'ca_id')),
    ('grid_losses', ('nd_id', 'ca_id')),
    ('erg_max', ('nd_id', 'ca_id', 'fl_id')),
    ('hyd_pwr_in_mt_max', ('pp_id',)),
    ('hyd_pwr_out_mt_min', ('pp_id',)),
    ('vc_fl', ('fl_id', 'nd_id')),
    ('co2_int', ('fl_id',)),
    ('nd_weight', ('nd_id',)),
    ('factor_lin_0', ('pp_id', 'ca_id')),
    ('factor_lin_1', ('pp_id', 'ca_id')),
    ('vc_om', ('pp_id', 'ca_id')),
    ('fc_om', ('pp_id', 'ca_id')),
    ('fc_dc', ('pp_id', 'ca_id')),
    ('fc_cp_ann', ('pp_id', 'ca_id')),
    ('ca_share_min', ('pp_id', 'ca_id')),
    ('ca_share_max', ('pp_id', 'ca_id')),
    ('pp_eff', ('pp_id', 'ca_id')),
    ('vc_ramp', ('pp_id', 'ca_id')),
    ('st_lss_hr', ('pp_id', 'ca_id')),
    ('st_lss_rt', ('pp_id', 'ca_id')),
    ('discharge_duration', ('pp_id', 'ca_id')),
#        ('hyd_erg_bc', ('sy', 'pp_id')), # this gets extremely large
    ('hyd_erg_min', ('pp_id',)),
    ('inflowprof', ('sy', 'pp_id', 'ca_id')),
    ('chpprof', ('sy', 'nd_id', 'ca_id')),
    ('supprof', ('sy', 'pp_id', 'ca_id')),
    ('pricesllprof', ('sy', 'pf_id')),
    ('pricebuyprof', ('sy', 'pf_id')),
    ('week_ror_output', ('wk', 'pp_id')),
    ('dmnd', ('sy', 'nd_id', 'ca_id')),
    ('erg_inp', ('nd_id', 'ca_id', 'fl_id')),
    ('erg_chp', ('pp_id', 'ca_id')),
    ('capchnge_max', tuple()),
    ]
dual = [('supply', ('sy', 'nd_id', 'ca_id'))]

list_collect = {'var_sy': var_sy,
                'var_mt': var_mt,
                'var_yr': var_yr,
                'var_tr': var_tr,
                'par': par,
                'dual': dual}

#def _make_dict(lst):
#    # can't use dict directly because of secondary table definition
#    # (value length 2)
#    return {comp[0]: comp[1] for comp in lst}
#
#
#dict_component_index = {**_make_dict(par), **_make_dict(var_mt),
#                        **_make_dict(var_sy), **_make_dict(var_tr),
#                        **_make_dict(var_yr)}
#
#dict_component_index = {key: dict_component_index[key] for key in sorted(dict_component_index)}

# table groups -> component -> index/secondary tb
DICT_TABLES = {lst: {tb[0]: tuple(tbb for tbb in tb[1:])
                     for tb in list_collect[lst]}
               for lst in list_collect}


# component -> table name
DICT_COMP_TABLE = {comp: (grp + '_' + spec[1]
                     if len(spec) > 1
                     else grp + '_' + comp)
              for grp, tbs in DICT_TABLES.items()
              for comp, spec in tbs.items()}

# component -> group
DICT_COMP_GROUP = {comp: grp
                  for grp, tbs in DICT_TABLES.items()
                  for comp, spec in tbs.items()}

# component -> index
DICT_COMP_IDX = {comp: spec[0]
            for grp, tbs in DICT_TABLES.items()
            for comp, spec in tbs.items()}

#DICT_COMP_IDX = {key: DICT_COMP_IDX[key] for key in sorted(DICT_COMP_IDX)}






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Mar 16 15:46:30 2019

@author: user
"""

import grimsel.auxiliary.sqlutils.aux_sql_func as aql
import grimsel_config as config

db = 'storage2'


sqlc = aql.SqlConnector(db, user=config.PSQL_USER,
                        password=config.PSQL_PASSWORD,
                        host=config.PSQL_HOST,
                        port=config.PSQL_PORT)



def yr_getter(par, data_type=False, rnge=range(2015, 2050 + 1, 5)):
    return [par + i if not data_type else (par + i, data_type)
            for i in [''] + ['_yr' + str(ii) for ii
            in rnge if not ii == 2015]]

def init_table(*args, **kwargs):
    con_cur = sqlc.get_pg_con_cur()
    print(con_cur, args, kwargs)
    return aql.init_table(*args, **kwargs, con_cur=con_cur)


def init_sql_tables(sc, db):

    tb_name = 'def_profile'
    cols = [('pf_id', 'SMALLINT'), ('pf', 'VARCHAR'), ('primary_nd', 'VARCHAR')]
    pk = ['pf_id']
    unique = ['pf']
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
               pk=pk, unique=unique, db=db)

    tb_name = 'def_pp_type'
    cols = [('pt_id',' SMALLINT'),
            ('pt',' varchar(20)'),
            ('pp_broad_cat', 'varchar(100)'),
            ('color', 'VARCHAR(7)')]
    pk = ['pt_id']
    unique = ['pt']
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'def_fuel'
    cols = [('fl_id', 'SMALLINT'), ('fl', 'varchar(20)'),
            ('co2_int', 'DOUBLE PRECISION'),
            ('is_ca', 'SMALLINT'),
            ('is_constrained', 'SMALLINT'),
            ('color', 'VARCHAR(7)')]
    pk = ['fl_id']
    unique = ['fl']
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'def_encar'
    cols = [('ca_id', 'SMALLINT'),
            ('fl_id', 'SMALLINT', sc + '.def_fuel(fl_id)'),
            ('ca', 'VARCHAR(2)')]
    pk = ['ca_id']
    unique = ['ca']
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'def_node'
    cols = [('nd_id', 'SMALLINT'),
            ('nd', 'VARCHAR(10)'),
            ('color', 'VARCHAR(7)')] + yr_getter('price_co2', 'DOUBLE PRECISION')
    pk = ['nd_id']
    unique = ['nd']
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'node_encar'
    cols = [('nd_id', 'SMALLINT', sc + '.def_node(nd_id)'),
            ('ca_id', 'SMALLINT', sc + '.def_encar(ca_id)'),
            ('dmnd_pf_id', 'SMALLINT', sc + '.def_profile(pf_id)'),
            ('grid_losses', 'DOUBLE PRECISION'),
            ('grid_losses_absolute', 'DOUBLE PRECISION'),
            ] + yr_getter('dmnd_sum', 'DOUBLE PRECISION')
    pk = ['nd_id', 'ca_id']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
               pk=pk, unique=unique, db=db)


    tb_name = 'def_month'
    cols = [('mt_id',' SMALLINT'),
            ('month_min_hoy',' SMALLINT'),
            ('month_weight',' SMALLINT'),
            ('mt',' VARCHAR(3)')]
    pk = ['mt_id']
    unique = ['name']
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'def_week'
    cols = [('wk_id',' SMALLINT'),
            ('wk',' SMALLINT'),
            ('week_weight', 'SMALLINT')]
    pk = ['wk_id']
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'def_plant'
    cols = [('pp_id',' SMALLINT'), ('pp',' VARCHAR(20)'),
            ('nd_id',' SMALLINT', sc + '.def_node(nd_id)'),
            ('fl_id',' SMALLINT', sc + '.def_fuel(fl_id)'),
            ('pt_id',' SMALLINT', sc + '.def_pp_type(pt_id)'),
            ('set_def_pr',' SMALLINT'),
            ('set_def_cain',' SMALLINT'),
            ('set_def_ror',' SMALLINT'),
            ('set_def_pp',' SMALLINT'), ('set_def_st',' SMALLINT'),
            ('set_def_hyrs',' SMALLINT'),
            ('set_def_chp',' SMALLINT'),
            ('set_def_add',' SMALLINT'),
            ('set_def_rem',' SMALLINT'),
            ('set_def_sll',' SMALLINT'),
            ('set_def_curt',' SMALLINT'),
            ('set_def_lin',' SMALLINT'),
            ('set_def_scen',' SMALLINT'),
            ('set_def_winsol',' SMALLINT'),
            ('set_def_tr', 'SMALLINT'),
            ('set_def_peak', 'SMALLINT')]
    pk = ['pp_id']
    unique = ['pp']
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'plant_month'
    cols = [('mt_id',' SMALLINT', sc + '.def_month(mt_id)'),
            ('pp_id',' SMALLINT', sc + '.def_plant(pp_id)'),
            ('hyd_erg_bc','DOUBLE PRECISION')]
    pk = ['mt_id', 'pp_id']
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'profsupply'
    cols = [('supply_pf_id',' SMALLINT', sc + '.def_profile(pf_id)'),
            ('hy', 'NUMERIC(6,2)'),
            ('value','NUMERIC(9,8)')]
    pk = ['supply_pf_id', 'hy']
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
               pk=pk, unique=unique, db=db)

    tb_name = 'plant_encar'
    cols = [('pp_id',' SMALLINT', sc + '.def_plant(pp_id)'),
            ('ca_id',' SMALLINT', sc + '.def_encar(ca_id)'),
            ('supply_pf_id', 'SMALLINT', sc + '.def_profile(pf_id)'),
            ('pp_eff','DOUBLE PRECISION'),
            ('erg_max','DOUBLE PRECISION'),
            ('discharge_duration','DOUBLE PRECISION'),
            ('st_lss_rt','DOUBLE PRECISION'),
            ('st_lss_hr','DOUBLE PRECISION'),
            ('factor_lin_0', 'DOUBLE PRECISION'),
            ('factor_lin_1','DOUBLE PRECISION'),
            ('cap_avlb', 'DOUBLE PRECISION'),
            ('vc_ramp','DOUBLE PRECISION'),
            ('vc_om','DOUBLE PRECISION'),
           ] + (yr_getter('cap_pwr_leg', 'DOUBLE PRECISION')
             +  yr_getter('erg_chp', 'DOUBLE PRECISION'))
    pk = ['pp_id', 'ca_id']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'plant_encar_scenarios'
    cols = [('pp_id',' SMALLINT', sc + '.def_plant(pp_id)'),
            ('ca_id',' SMALLINT', sc + '.def_encar(ca_id)'),
            ('scenario', 'VARCHAR'),
           ] + (yr_getter('cap_pwr_leg', 'DOUBLE PRECISION'))
    pk = ['pp_id', 'ca_id', 'scenario']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'imex_comp'
    cols = [('nd_id', 'SMALLINT', sc + '.def_node(nd_id)'),
            ('nd_2_id', 'SMALLINT', sc + '.def_node(nd_id)'),
            ] + yr_getter('erg_trm', 'DOUBLE PRECISION', [2015])
    pk = ['nd_id', 'nd_2_id']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'profdmnd'
    cols = [('dmnd_pf_id', 'SMALLINT', sc + '.def_profile(pf_id)'),
            ('hy', 'NUMERIC(6,2)')] + yr_getter('value', 'NUMERIC(18,9)', [2015])
    pk = ['hy', 'dmnd_pf_id']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'profchp'
    cols = [('nd_id', 'SMALLINT', sc + '.def_node(nd_id)'),
            ('ca_id', 'SMALLINT', sc + '.def_encar(ca_id)'),
            ('hy', 'SMALLINT'), ('value', 'DOUBLE PRECISION')]
    pk = ['hy', 'nd_id']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'profinflow'
    cols = [('pp_id', 'SMALLINT', sc + '.def_plant(pp_id)'),
            ('ca_id', 'SMALLINT', sc + '.def_encar(ca_id)'),
            ('hy', 'SMALLINT'), ('value', 'DOUBLE PRECISION')]
    pk = ['hy', 'pp_id']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'profprice'
    cols = [('hy', 'NUMERIC(6,2)'),
            ('price_pf_id', 'SMALLINT', sc + '.def_profile(pf_id)'),
           ] + yr_getter('value', 'DOUBLE PRECISION', [2015])
    pk = ['hy', 'price_pf_id']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'fuel_node_encar'
    cols = ([('fl_id', 'SMALLINT', sc + '.def_fuel(fl_id)'),
             ('nd_id', 'SMALLINT', sc + '.def_node(nd_id)'),
             ('ca_id', 'SMALLINT', sc + '.def_encar(ca_id)'),
             ('pricesll_pf_id', 'SMALLINT', sc + '.def_profile(pf_id)'),
             ('pricebuy_pf_id', 'SMALLINT', sc + '.def_profile(pf_id)'),
             ('is_chp', 'SMALLINT'),
             ] + yr_getter('erg_inp', 'DOUBLE PRECISION')
               + yr_getter('vc_fl', 'DOUBLE PRECISION'))
    pk = ['fl_id', 'nd_id']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
               pk=pk, unique=unique, db=db)

    tb_name = 'fuel_node_encar_scenarios'
    cols = ([('fl_id', 'SMALLINT', sc + '.def_fuel(fl_id)'),
             ('nd_id', 'SMALLINT', sc + '.def_node(nd_id)'),
             ('ca_id', 'SMALLINT', sc + '.def_encar(ca_id)'),
             ('scenario', 'VARCHAR'),
             ] + yr_getter('erg_inp', 'DOUBLE PRECISION'))
    pk = ['fl_id', 'nd_id', 'scenario']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    # table with monthly parameter modifiers
    tb_name = 'parameter_month'
    cols = ([('set_1_name', 'VARCHAR'), # from {'nd_id', 'fl_id', 'pp_id'}
             ('set_2_name', 'VARCHAR'), # from {'nd_id', 'fl_id', 'pp_id'}
             ('set_1_id', 'SMALLINT'),
             ('set_2_id', 'SMALLINT'),
             ('mt_id',' SMALLINT', sc + '.def_month(mt_id)'),
             ('parameter', 'VARCHAR'), # the parameter this applies to
             ('mt_fact', 'NUMERIC(10,9)'),
             ('mt_fact_others', 'NUMERIC(10,9)'),
             ])
    pk = ['parameter', 'set_1_id', 'set_2_id', 'mt_id']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)

    tb_name = 'node_connect'
    cols = [('nd_id', 'SMALLINT', sc + '.def_node (nd_id)'),
            ('nd_2_id', 'SMALLINT', sc + '.def_node (nd_id)'),
            ('ca_id', 'SMALLINT', sc + '.def_encar (ca_id)'),
            ('mt_id', 'SMALLINT', sc + '.def_month(mt_id)'),
            ('cap_trme_leg', 'DOUBLE PRECISION'),
            ('cap_trmi_leg', 'DOUBLE PRECISION'),
            ]
    pk = ['nd_id', 'nd_2_id', 'mt_id']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)


    tb_name = 'hydro'
    cols = [('pp_id',' SMALLINT', sc + '.def_plant(pp_id)'),
            ('min_erg_mt_out_share', 'DOUBLE PRECISION'),
            ('max_erg_mt_in_share', 'DOUBLE PRECISION'),
            ('min_erg_share', 'DOUBLE PRECISION')]
    pk = ['pp_id']
    unique = []
    init_table(tb_name=tb_name, cols=cols, schema=sc, ref_schema=sc,
                   pk=pk, unique=unique, db=db)
